from __future__ import annotations

from datetime import datetime
from unittest.mock import patch

import pytest
from yardstick import artifact


def test_sort_matches():
    a = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2019-18276"),
        package=artifact.Package(name="bash", version="5.0-6ubuntu1.1"),
    )

    b = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2016-2781"),
        package=artifact.Package(name="coreutils", version="8.30-3ubuntu2"),
        config=artifact.ScanConfiguration(
            image_repo="fff",
            image_digest="fff",
            tool_name="fff",
            tool_version="fff",
        ),
    )

    c = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2016-10228"),
        package=artifact.Package(name="libc-bin", version="2.31-0ubuntu9.2"),
        config=artifact.ScanConfiguration(
            image_repo="ddd",
            image_digest="ddd",
            tool_name="ddd",
            tool_version="ddd",
        ),
    )

    d = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2018-20839"),
        package=artifact.Package(name="libsystemd0", version="245.4-4ubuntu3.2"),
    )

    e = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2018-299999999"),
        package=artifact.Package(name="libsystemd0", version="25"),
        config=artifact.ScanConfiguration(
            image_repo="a",
            image_digest="a",
            tool_name="a",
            tool_version="a",
        ),
    )

    assert sorted([d, e, c, b, a]) == [a, b, c, d, e]


def test_sort_packages():
    a = artifact.Package(name="bash", version="5.0-6ubuntu1.1")

    b = artifact.Package(name="coreutils", version="8.30-3ubuntu2")

    c = artifact.Package(name="libc-bin", version="2.31-0ubuntu9.2")

    d = artifact.Package(name="libsystemd0", version="245.4-4ubuntu3.2")

    e = artifact.Package(name="libsystemd0", version="25")

    assert sorted([d, c, e, b, a]) == [a, b, c, d, e]


class TestImageSpecifier:
    @pytest.mark.parametrize(
        ("specifier", "image", "expected"),
        [
            # exact match
            (
                artifact.ImageSpecifier(
                    exact="something/exact:latest",
                ),
                "something/exact:latest",
                True,
            ),
            # missing tail
            (
                artifact.ImageSpecifier(
                    exact="something/exact:lates",
                ),
                "something/exact:latest",
                False,
            ),
            # matches prefix
            (
                artifact.ImageSpecifier(
                    prefix="something/exact:lates",
                ),
                "something/exact:latest",
                True,
            ),
            # matches regex
            (
                artifact.ImageSpecifier(
                    regex="^.*/exact:latest$",
                ),
                "something/exact:latest",
                True,
            ),
        ],
    )
    def test_matches(self, specifier, image, expected):
        assert expected == specifier.matches_image(image)


def test_tool():
    t = artifact.Tool("grype@main")

    assert t.name == "grype"
    assert t.version == "main"
    assert t.id == "grype"

    t = artifact.Tool("grype@main", label="frank")

    assert t.name == "grype"
    assert t.version == "main"
    assert t.id == "grype[frank]"


def test_image():
    i = artifact.Image("docker.io/place/ubuntu:thing@sha256:123")

    assert i.repository == "docker.io/place/ubuntu"
    assert i.tag == "thing"
    assert i.digest == "sha256:123"
    assert i.repository_encoded == "docker.io+place+ubuntu"
    assert i.encoded == "docker.io+place+ubuntu@sha256:123"

    assert i.is_like(other="docker.io/place/ubuntu@sha256:123")
    assert not i.is_like(other="docker.io/place/ubuntu:SOMETHINGELSE@sha256:123")
    assert not i.is_like(other="docker.io/SOMETHINGELSE/ubuntu@sha256:123")
    assert not i.is_like(other="docker.io/place/ubuntu@sha256:SOMETHINGELSE")


def test_scan_configuration():
    ts_rfc3339 = "2022-09-06T16:07:01.138937+00:00"
    ts = datetime.fromisoformat(ts_rfc3339)

    s = artifact.ScanConfiguration(
        image_repo="docker.io+place+ubuntu",
        image_digest="sha256:123",
        image_tag="stuff",
        tool_name="grype",
        tool_version="main",
        timestamp=ts,
    )

    assert s.image_repo == "docker.io/place/ubuntu"
    assert s.image_digest == "sha256:123"
    assert s.image_tag == "stuff"
    assert s.tool_name == "grype"
    assert s.tool_version == "main"
    assert s.tool_input is None
    assert s.timestamp == datetime.fromisoformat(ts_rfc3339)
    assert s.timestamp_rfc3339 == ts_rfc3339
    assert s.tool == "grype@main"
    assert s.image == "docker.io/place/ubuntu@sha256:123"
    assert s.image_encoded == "docker.io+place+ubuntu@sha256:123"
    assert s.image_repo_encoded == "docker.io+place+ubuntu"
    assert s.path == "docker.io/place/ubuntu@sha256:123/grype@main/2022-09-06T16:07:01.138937+00:00"
    assert s.encoded_path == "docker.io+place+ubuntu@sha256:123/grype@main/2022-09-06T16:07:01.138937+00:00"


def test_dt_encoder():
    ts = datetime.fromisoformat("2022-09-06T16:07:01+00:00")
    assert artifact.DTEncoder().default(ts) == "2022-09-06T16:07:01+00:00"

    ts = datetime.fromisoformat("2022-09-06T16:07:01.133789+00:00")
    assert artifact.DTEncoder().default(ts) == "2022-09-06T16:07:01.133789+00:00"


@pytest.mark.parametrize(
    ("subject_id", "subject_cve_id", "expected", "year_from_cve_only", "cve_id_lookup"),
    [
        # below cases prefer the primary ID year over the CVE year
        pytest.param(
            "CVE-2018-20839",
            None,
            2018,
            False,
            None,  # no-op
            id="cve as ID",
        ),
        pytest.param(
            "CVE-2018-20839",
            "CVE-2000-1234567",
            2018,
            False,
            None,  # no-op
            id="cve as ID with CVE override (INVALID)",
        ),
        pytest.param(
            "ALASKERNEL-5.1-2022-0002",
            None,
            2022,
            False,
            None,  # no-op
            id="ALASKERNEL",
        ),
        pytest.param(
            "ALASKERNEL-5.1-2022-0002",
            "CVE-2000-1234567",
            2022,
            False,
            None,  # no-op
            id="ALASKERNEL with CVE (ignore)",
        ),
        pytest.param(
            "ALAS-2023-0003",
            None,
            2023,
            False,
            None,  # no-op
            id="ALAS",
        ),
        pytest.param(
            "ALAS-2023-0003",
            "CVE-2000-1234567",
            2023,
            False,
            None,  # no-op
            id="ALAS with CVE (ignore)",
        ),
        pytest.param(
            "ELSA-2023-0003",
            None,
            2023,
            False,
            None,  # no-op
            id="ELSA",
        ),
        pytest.param(
            "ELSA-2023-0003",
            "CVE-2000-1234567",
            2023,
            False,
            None,  # no-op
            id="ELSA with CVE (ignore)",
        ),
        # below cases prefer CVE over primary ID...
        pytest.param(
            "CVE-2018-20839",
            "CVE-2000-1234567",
            2000,
            True,
            None,  # no-op
            id="cve as ID with CVE override (INVALID) -- use CVE when present",
        ),
        pytest.param(
            "ALASKERNEL-5.1-2022-0002",
            None,
            None,
            True,
            None,  # no-op
            id="ALASKERNEL -- use CVE when present (missing)",
        ),
        pytest.param(
            "ALASKERNEL-5.1-2022-0002",
            None,
            2000,
            True,
            "CVE-2000-1234567",
            id="ALASKERNEL -- use CVE when present (lookup returns CVE)",
        ),
        pytest.param(
            "ALASKERNEL-5.1-2022-0002",
            "CVE-2000-1234567",
            2000,
            True,
            None,  # no-op
            id="ALASKERNEL with CVE (ignore) -- use CVE when present",
        ),
        pytest.param(
            "ALAS-2023-0003",
            None,
            None,
            True,
            None,  # no-op
            id="ALAS -- use CVE when present (missing)",
        ),
        pytest.param(
            "ALAS-2023-0003",
            None,
            2000,
            True,
            "CVE-2000-1234567",
            id="ALAS -- use CVE when present (lookup returns CVE)",
        ),
        pytest.param(
            "ALAS-2023-0003",
            "CVE-2000-1234567",
            2000,
            True,
            None,  # no-op
            id="ALAS with CVE (ignore) -- use CVE when present",
        ),
        pytest.param(
            "ELSA-2023-0003",
            None,
            None,
            True,
            None,  # no-op
            id="ELSA -- use CVE when present (missing)",
        ),
        pytest.param(
            "ELSA-2023-0003",
            None,
            2000,
            True,
            "CVE-2000-1234567",
            id="ELSA -- use CVE when present (lookup returns CVE)",
        ),
        pytest.param(
            "ELSA-2023-0003",
            "CVE-2000-1234567",
            2000,
            True,
            None,  # no-op
            id="ELSA with CVE (ignore) -- use CVE when present",
        ),
    ],
)
def test_effective_year(
    subject_id: str,
    subject_cve_id: str | None,
    expected: int | None,
    year_from_cve_only: bool,
    cve_id_lookup: str | None,
):
    # first, check that the vulnerability object is alright
    vuln = artifact.Vulnerability(id=subject_id, cve_id=subject_cve_id)

    with patch("yardstick.utils.grype_db.normalize_to_cve", lambda _: cve_id_lookup):
        assert expected == vuln.effective_year(by_cve=year_from_cve_only)

    # now check that the label entry behaves the same
    with patch("yardstick.utils.grype_db.normalize_to_cve", lambda _: subject_cve_id):
        le = artifact.LabelEntry(
            image="dontcare",
            vulnerability_id=subject_id,
            label=artifact.Label.TruePositive,
            lookup_effective_cve=True,
        )

    with patch("yardstick.utils.grype_db.normalize_to_cve", lambda _: cve_id_lookup):
        assert expected == le.effective_year(by_cve=year_from_cve_only)



import re
import sys

import click
from tabulate import tabulate

import yardstick
from yardstick import store
from yardstick import validate as val
from yardstick.cli import config, display
from yardstick.validate import Gate, GateInputDescription


class bcolors:
    HEADER: str = "\033[95m"
    OKBLUE: str = "\033[94m"
    OKCYAN: str = "\033[96m"
    OKGREEN: str = "\033[92m"
    WARNING: str = "\033[93m"
    FAIL: str = "\033[91m"
    BOLD: str = "\033[1m"
    UNDERLINE: str = "\033[4m"
    RESET: str = "\033[0m"


if not sys.stdout.isatty():
    bcolors.HEADER = ""
    bcolors.OKBLUE = ""
    bcolors.OKCYAN = ""
    bcolors.OKGREEN = ""
    bcolors.WARNING = ""
    bcolors.FAIL = ""
    bcolors.BOLD = ""
    bcolors.UNDERLINE = ""
    bcolors.RESET = ""


@click.command()
@click.pass_obj
@click.option(
    "--image",
    "-i",
    "images",
    multiple=True,
    help="filter down to one or more images to validate with (don't use the full result set)",
)
@click.option(
    "--label-comparison",
    "-l",
    "always_run_label_comparison",
    is_flag=True,
    help="run label comparison irregardless of relative comparison results",
)
@click.option(
    "--breakdown-by-ecosystem",
    "-e",
    is_flag=True,
    help="show label comparison results broken down by ecosystem",
)
@click.option("--verbose", "-v", "verbosity", count=True, help="show details of all comparisons")
@click.option(
    "--result-set",
    "-r",
    "result_sets",
    multiple=True,
    default=[],
    help="the result set to use for the quality gate",
)
@click.option(
    "--all",
    "all_result_sets",
    is_flag=True,
    default=False,
    help="validate all known result sets",
)
def validate(
    cfg: config.Application,
    images: list[str],
    always_run_label_comparison: bool,
    breakdown_by_ecosystem: bool,
    verbosity: int,
    result_sets: list[str],
    all_result_sets: bool,
):
    # TODO: don't artificially inflate logging; figure out what to print
    setup_logging(verbosity + 3)
    if all_result_sets and result_sets and len(result_sets) > 0:  # default result set will be present anyway
        raise ValueError(f"cannot pass --all and -r / --result-set: {all_result_sets} {result_sets}")

    if all_result_sets:
        result_sets = [r for r in cfg.result_sets.keys()]

    if not result_sets:
        raise ValueError("must pass --result-set / -r at least once or --all to validate all result sets")

    # let's not load any more labels than we need to, base this off of the images we're validating
    if not images:
        unique_images = set()
        for r in result_sets:
            result_set_obj = store.result_set.load(name=r)
            for state in result_set_obj.state:
                if state and state.config and state.config.image:
                    unique_images.add(state.config.image)
        images = sorted(list(unique_images))

    click.echo("Loading label entries...", nl=False)
    label_entries = store.labels.load_for_image(images, year_max_limit=cfg.max_year_for_any_result_set(result_sets))
    click.echo(f"done! {len(label_entries)} entries loaded")

    gates = []
    for result_set in result_sets:
        rs_config = cfg.result_sets[result_set]
        for gate_config in rs_config.validations:
            if gate_config.max_year is None:
                gate_config.max_year = cfg.default_max_year

            click.echo(f"{bcolors.HEADER}{bcolors.BOLD}Validating with {result_set!r}{bcolors.RESET}")
            new_gates = val.validate_result_set(
                gate_config,
                result_set,
                images=images,
                always_run_label_comparison=always_run_label_comparison,
                verbosity=verbosity,
                label_entries=label_entries,
            )
            for gate in new_gates:
                show_results_for_image(gate.input_description, gate)

            gates.extend(new_gates)
        click.echo()

        if breakdown_by_ecosystem:
            click.echo(
                f"{bcolors.HEADER}Breaking down label comparison by ecosystem performance...{bcolors.RESET}",
            )
            results_by_image, label_entries, stats = yardstick.compare_results_against_labels_by_ecosystem(
                result_set=result_set,
                year_max_limit=cfg.max_year_for_result_set(result_set),
                label_entries=label_entries,
            )
            display.labels_by_ecosystem_comparison(
                results_by_image,
                stats,
                show_images_used=False,
            )
            click.echo()

    failure = not all([gate.passed() for gate in gates])
    if failure:
        click.echo("Reasons for quality gate failure:")
    for gate in gates:
        for reason in gate.reasons:
            click.echo(f"   - {reason} ({gate.input_description.image})")

    if failure:
        click.echo()
        click.echo(f"{bcolors.FAIL}{bcolors.BOLD}Quality gate FAILED{bcolors.RESET}")
        sys.exit(1)
    else:
        click.echo(f"{bcolors.OKGREEN}{bcolors.BOLD}Quality gate passed!{bcolors.RESET}")


def setup_logging(verbosity: int):
    # pylint: disable=redefined-outer-name, import-outside-toplevel
    import logging.config

    if verbosity in [0, 1, 2]:
        log_level = "WARN"
    elif verbosity == 3:
        log_level = "INFO"
    else:
        log_level = "DEBUG"

    logging.config.dictConfig(
        {
            "version": 1,
            "formatters": {
                "standard": {
                    # [%(module)s.%(funcName)s]
                    "format": "%(asctime)s [%(levelname)s] %(message)s",
                    "datefmt": "",
                },
            },
            "handlers": {
                "default": {
                    "level": log_level,
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stderr",
                },
            },
            "loggers": {
                "": {  # root logger
                    "handlers": ["default"],
                    "level": log_level,
                },
            },
        }
    )


def show_delta_commentary(gate: Gate):
    if not gate.deltas:
        click.echo("No differences found between tooling (with labels)")
        return

    header_row = ["TOOL PARTITION", "PACKAGE", "VULNERABILITY", "LABEL", "COMMENTARY"]

    all_rows = []
    for delta in gate.deltas:
        color = ""
        if delta.is_improved:
            color = bcolors.OKBLUE
        elif delta.is_improved is not None and not delta.is_improved:
            color = bcolors.FAIL
        all_rows.append(
            [
                f"{color}{delta.tool} ONLY{bcolors.RESET}",
                f"{color}{delta.package_name}@{delta.package_version}{bcolors.RESET}",
                f"{color}{delta.vulnerability_id}{bcolors.RESET}",
                f"{color}{delta.label}{bcolors.RESET}",
                f"{delta.commentary}",
            ]
        )

    def escape_ansi(line):
        ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
        return ansi_escape.sub("", line)

    # sort but don't consider ansi escape codes
    all_rows = sorted(all_rows, key=lambda x: escape_ansi(str(x[0] + x[1] + x[2] + x[3])))
    click.echo("Match differences between tooling (with labels):")
    indent = "   "
    click.echo(
        indent
        + tabulate(
            [header_row] + all_rows,
            tablefmt="plain",
        ).replace("\n", "\n" + indent)
        + "\n"
    )


def show_results_for_image(input_description: GateInputDescription, gate: Gate):
    click.echo(f"   Results used for image {input_description.image}:")
    for idx, description in enumerate(input_description.configs):
        branch = "├──"
        if idx == len(input_description.configs) - 1:
            branch = "└──"
        label = " "
        if description.tool_label and len(description.tool_label) > 0:
            label = f" ({description.tool_label}) "
        click.echo(f"    {branch} {description.id} : {description.tool}{label} against {input_description.image}")
    if gate.deltas:
        click.echo(f"Deltas for {input_description.image}:")
        show_delta_commentary(gate)
    click.echo("-" * 80)



import abc
from typing import List, Optional

from yardstick import artifact


class SBOMGenerator(abc.ABC):
    @staticmethod
    @abc.abstractmethod
    def install(version: str, path: Optional[str] = None, **kwargs) -> "SBOMGenerator":
        raise NotImplementedError

    @abc.abstractmethod
    def capture(self, image: str, tool_input: Optional[str]) -> str:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def parse(
        result: str,
        config: artifact.ScanConfiguration,
    ) -> List[artifact.Package]:
        raise NotImplementedError



from __future__ import annotations

import json
import logging
import os
import shutil

from yardstick import artifact
from yardstick.store import config as store_config
from yardstick.store import scan_result, tool


def _store_root(store_root: str | None = None):
    if not store_root:
        store_root = store_config.get().store_root

    return tool.result_set_path(store_root=store_root)


def clear(store_root: str | None = None):
    shutil.rmtree(_store_root(store_root=store_root), ignore_errors=True)


def store_paths(name: str, store_root: str | None = None) -> str:
    parent_dir = _store_root(store_root=store_root)

    return os.path.join(parent_dir, f"{name}.json")


def save(results: artifact.ResultSet, store_root: str | None = None):
    if not isinstance(results, artifact.ResultSet):
        raise RuntimeError(f"only ResultSet is supported, given {type(results)}")

    path = store_paths(results.name, store_root=store_root)
    logging.debug(f"storing result set state {results.name!r}")

    os.makedirs(os.path.dirname(path), exist_ok=True)

    with open(path, "w", encoding="utf-8") as data_file:
        data_file.write(json.dumps(results.to_dict(), indent=2))  # type: ignore[attr-defined]


def load(name: str, store_root: str | None = None) -> artifact.ResultSet:
    data_path = store_paths(name, store_root=store_root)
    logging.debug(f"loading result set {name!r} location={data_path!r}")

    with open(data_path, encoding="utf-8") as data_file:
        data_json = data_file.read()
    return artifact.ResultSet.from_json(data_json)  # type: ignore[attr-defined]


def load_scan_results(
    name: str,
    year_max_limit: int | None = None,
    store_root: str | None = None,
    skip_sbom_results: bool = False,
) -> list[artifact.ScanResult]:
    data_path = store_paths(name, store_root=store_root)
    logging.debug(
        f"loading scan results from result set {name!r} location={data_path!r}",
    )

    result_set = load(name, store_root=store_root)

    descriptions = [result_state.config.path for result_state in result_set.state if result_state.config]
    return scan_result.load_by_descriptions(
        descriptions,
        year_max_limit=year_max_limit,
        store_root=store_root,
        skip_sbom_results=skip_sbom_results,
    )


def exists(name: str, store_root: str | None = None) -> bool:
    data_path = store_paths(name, store_root=store_root)
    return os.path.exists(data_path)


def load_all(store_root: str | None = None) -> list[artifact.ResultSet]:
    parent_dir = _store_root(store_root=store_root)
    result_sets = []

    for file_name in os.listdir(parent_dir):
        if file_name.endswith(".json"):
            result_sets.append(load(file_name[:-5], store_root=store_root))

    return result_sets



import atexit
import functools
import json
import logging
import os
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
from typing import Dict, List, Optional

import git

from yardstick import artifact, utils
from yardstick.tool.sbom_generator import SBOMGenerator
from yardstick.utils import github


class Syft(SBOMGenerator):
    _latest_version_from_github: Optional[str] = None

    def __init__(
        self,
        path: str,
        version_detail: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
    ):
        self.path = path
        self._env = env
        if version_detail:
            self.version_detail = version_detail

    @staticmethod
    @functools.cache
    def latest_version_from_github():
        return github.get_latest_release_version(project="syft")

    @staticmethod
    def _install_from_installer(
        version: str,
        path: Optional[str] = None,
        use_cache: Optional[bool] = True,
        **kwargs,  # noqa: ARG004
    ) -> "Syft":
        logging.debug(f"installing syft version={version!r} from installer")
        tool_exists = False

        if not use_cache and path:
            shutil.rmtree(path, ignore_errors=True)

        if not path:
            path = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, path)

        os.makedirs(path, exist_ok=True)
        if os.path.exists(os.path.join(path, "syft")):
            tool_exists = True

        if not tool_exists:
            subprocess.check_call(
                [
                    "sh",
                    "-c",
                    f"curl -sSfL https://raw.githubusercontent.com/anchore/syft/main/install.sh | sh -s -- -b {path} {version}",
                ],
            )

            # note on S103 exception: syft must be executable and it is not valid to assume that
            # a single user or group should be restricted to using it. 0755 is the default
            # permission used for system bin/* contents.
            os.chmod(f"{path}/syft", 0o755)  # noqa: S103
        else:
            logging.debug(f"using existing syft installation {path!r}")

        return Syft(path=path, version_detail=version)

    @classmethod
    def _install_from_git(  # noqa: PLR0912
        cls,
        version: str,
        path: Optional[str] = None,
        use_cache: Optional[bool] = True,
        **kwargs,  # noqa: ARG003
    ) -> "Syft":
        logging.debug(f"installing syft version={version!r} from git")
        tool_exists = False

        repo_url = "github.com/anchore/syft"
        if version.startswith("github.com"):
            repo_url = version
            if "@" in version:
                version = version.split("@")[1]
                repo_url = repo_url.split("@")[0]
            else:
                version = "main"

        if not use_cache and path:
            shutil.rmtree(path, ignore_errors=True)

        if not path:
            path = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, path)

        # grab the latest source code into a local directory
        repo_path = os.path.join(path, "source")

        if os.path.exists(repo_path):
            logging.debug(f"found existing syft repo at {repo_path!r}")
            # use existing source
            try:
                repo = git.Repo(repo_path)
            except:
                logging.error(f"failed to open existing syft repo at {repo_path!r}")
                raise
        else:
            logging.debug("cloning the syft git repo: {repo_url!r}")
            # clone the repo
            os.makedirs(repo_path)
            repo = git.Repo.clone_from("https://{repo_url}.git", repo_path)

        # checkout the ref in question
        repo.git.fetch("origin", version)
        repo.git.checkout(version)

        # get userful ref
        description = repo.git.describe("--tags", "--always", "--long")
        path = os.path.join(path, "git_install", description)

        logging.debug(f"found syft git description={description!r}")

        if os.path.exists(os.path.join(path, "syft")):
            tool_exists = True
        else:
            os.makedirs(path)

        abspath = os.path.abspath(path)
        if not tool_exists:
            logging.debug(f"installing syft to {abspath!r}")
            cls._run_go_build(
                abs_install_dir=abspath,
                repo_path=repo_path,
                description=description,
                binpath=path,
            )
        else:
            logging.debug(f"using existing syft installation {abspath!r}")

        return Syft(path=path, version_detail=description)

    @classmethod
    def _install_from_path(
        cls,
        path: Optional[str],
        src_path: str,
    ) -> "Syft":
        # get the description and head ref from the repo
        src_repo_path = os.path.abspath(os.path.expanduser(src_path))
        build_version = utils.local_build_version_suffix(src_repo_path)
        logging.debug(f"installing syft from path={src_repo_path!r}")
        logging.debug(f"installing syft to path={path!r}")
        if not path:
            path = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, path)
        dest_path = os.path.join(
            path.replace("path:", ""),
            build_version,
            "local_install",
        )
        os.makedirs(dest_path, exist_ok=True)
        cls._run_go_build(
            abs_install_dir=os.path.abspath(dest_path),
            description=f"{path}:{build_version}",
            repo_path=src_repo_path,
            binpath=dest_path,
        )

        return Syft(path=dest_path)

    @staticmethod
    def _run_go_build(
        abs_install_dir: str,
        repo_path: str,
        description: str,
        binpath: str,
        version_ref: str = "github.com/anchore/syft/internal/version.version",
    ):
        logging.debug(f"installing syft via build to {abs_install_dir!r}")

        main_pkg_path = "./cmd/syft"
        if not os.path.exists(os.path.join(repo_path, "cmd", "syft", "main.go")):
            # support legacy installations, when the main.go was in the root of the repo
            main_pkg_path = "."

        c = f"go build -ldflags \"-w -s -extldflags '-static' -X {version_ref}={description}\" -o {abs_install_dir} {main_pkg_path}"
        logging.debug(f"running {c!r}")

        e = {"GOBIN": abs_install_dir, "CGO_ENABLED": "0"}
        e.update(os.environ)

        subprocess.check_call(
            shlex.split(c),
            stdout=sys.stdout,
            stderr=sys.stderr,
            cwd=repo_path,
            env=e,
        )

        # note on S103 exception: syft must be executable and it is not valid to assume that
        # a single user or group should be restricted to using it. 0755 is the default
        # permission used for system bin/* contents.
        os.chmod(f"{binpath}/syft", 0o755)  # noqa: S103

    @classmethod
    def install(
        cls,
        version: str,
        path: Optional[str] = None,
        use_cache: Optional[bool] = True,
        **kwargs,
    ) -> "Syft":
        if version == "latest":
            version = cls.latest_version_from_github()
            if path:
                path = os.path.join(os.path.dirname(path), version)
            logging.info(f"latest syft release found is {version}")

        # check if the version is a semver...
        if re.match(
            r"^v(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$",
            version,
        ):
            tool_obj = cls._install_from_installer(
                version=version,
                path=path,
                use_cache=use_cache,
                **kwargs,
            )
        elif version.startswith("path:"):
            tool_obj = cls._install_from_path(
                path=path,
                src_path=version.removeprefix("path:"),
            )
        else:
            tool_obj = cls._install_from_git(
                version=version,
                path=path,
                use_cache=use_cache,
                **kwargs,
            )

        return tool_obj

    def env(self):
        env = os.environ.copy()
        env["SYFT_CHECK_FOR_APP_UPDATE"] = "false"
        if self._env:
            env.update(self._env)
        return env

    @staticmethod
    def parse(
        result: str,
        config: artifact.ScanConfiguration,
    ) -> List[artifact.Package]:
        logging.debug("parsing syft results")

        results: List[artifact.Package] = []
        obj = json.loads(result)

        # patch the config with the digest found in the results
        repo = config.image.split(":")[0]
        for digest in utils.dig(obj, "source", "target", "repoDigests", default=[]):
            if digest.startswith(repo + "@"):
                config.image_digest = digest.split("@")[1]
                break

        for entry in obj["artifacts"]:
            # TODO: normalize version here
            pkg = artifact.Package(name=entry["name"], version=entry["version"])
            results.append(pkg)
        return results

    def capture(self, image: str, tool_input: Optional[str]) -> str:
        i = image if tool_input is None else tool_input
        logging.debug(f"running syft with input={i}")
        return self.run("-o", "json", i)

    def run(self, *args) -> str:
        return subprocess.check_output(
            [f"{self.path}/syft", *args],
            env=self.env(),
        ).decode("utf-8")



from __future__ import annotations

import copy
import glob
import itertools
import json
import logging
import os
import shutil
from collections import defaultdict

from yardstick import artifact
from yardstick.store import config as store_config
from yardstick.store import naming, tool
from yardstick.tool import sbom_generator, tools, vulnerability_scanner


def _store_root(store_root: str | None = None):
    if not store_root:
        store_root = store_config.get().store_root

    return tool.results_path(store_root=store_root)


def clear(store_root: str | None = None):
    shutil.rmtree(_store_root(store_root=store_root), ignore_errors=True)


def store_paths(
    config: artifact.ScanConfiguration,
    suffix: str = naming.SUFFIX,
    store_root: str | None = None,
) -> tuple[str, str]:
    # repo@digest/tool@version/timestamp/data.json
    # repo@digest/tool@version/timestamp/metadata.json

    parent_dir = _store_root(store_root=store_root)
    result_dir = os.path.join(
        parent_dir,
        f"{config.image_encoded}",
        f"{config.tool_name.replace('/', '_')}@{config.tool_version.replace('/', '_')}",
        config.timestamp_rfc3339,
    )

    return os.path.join(result_dir, "data" + suffix), os.path.join(
        result_dir,
        "metadata" + suffix,
    )


# note: we intentionally split up the data and metadata such that matches are recomputed with the latest code changes
def save(raw: str, results: artifact.ScanResult, store_root: str | None = None):
    if not isinstance(results, artifact.ScanResult):
        raise RuntimeError(f"only ScanResult is supported, given {type(results)}")

    data_path, metadata_path = store_paths(results.config, store_root=store_root)
    logging.debug(f"storing result config={results.config!r} location={data_path!r}")

    os.makedirs(os.path.dirname(data_path), exist_ok=True)
    os.makedirs(os.path.dirname(metadata_path), exist_ok=True)

    with open(data_path, "w", encoding="utf-8") as data_file:
        data_file.write(raw)

    results_dict = results.to_dict()  # type: ignore[attr-defined]
    if "matches" in results_dict:
        del results_dict["matches"]
    if "packages" in results_dict:
        del results_dict["packages"]

    with open(metadata_path, "w", encoding="utf-8") as metadata_file:
        metadata_file.write(json.dumps(results_dict, cls=artifact.DTEncoder))


def find(
    by_image: str = "",
    by_tool: str = "",
    by_time: str = "",
    by_description: str = "",
    store_root: str | None = None,
) -> list[artifact.ScanConfiguration]:
    json_path = tool.results_path(store_root=store_root)

    image_spec = "*"
    tool_spec = "*"
    time_spec = "*"

    if by_image:
        img = artifact.Image(by_image)
        if not img.digest:
            raise RuntimeError(f"image {by_image} requires a digest to search by image")
        image_spec = f"{img.repository_encoded}@{img.digest}"

    if by_tool:
        tool_spec = by_tool if "@" in by_tool else f"{by_tool}@*"

    if by_time and by_time != "latest":
        time_spec = by_time

    is_id = "/" not in by_description and by_description

    if by_description and by_description.count("/") >= 2:
        tool_name_side, tool_version_side = by_description.rsplit("@", 1)
        tool_name_side_fields = tool_name_side.rsplit("/", 1)
        tool_name = tool_name_side_fields[-1]
        repos = tool_name_side_fields[0]
        image_spec = repos.replace("/", "+")
        tool_version_side_fields = tool_version_side.rsplit("/", 1)
        tool_version = tool_version_side_fields[0]
        tool_spec = f"{tool_name}@{tool_version}"
        time_spec = tool_version_side_fields[-1]

        # to account for lables with [], which should be escaped
        tool_spec = glob.escape(tool_spec)

    search_tuple = f"{image_spec}/{tool_spec.replace('/', '_')}/{time_spec}"

    results = defaultdict(list)

    glob_str = f"{json_path}/{search_tuple}/metadata.json"

    logging.debug(f"searching for {glob_str}")

    for metadata_file in glob.glob(glob_str):
        image_tool_dir = os.path.dirname(os.path.dirname(metadata_file))
        with open(metadata_file, encoding="utf-8") as fd:
            metadata_dict = json.load(fd)
            cfg = artifact.ScanConfiguration.from_dict(  # type: ignore[attr-defined]
                metadata_dict["config"],
            )

            if is_id and by_description != cfg.ID:
                continue

            results[image_tool_dir].append(cfg)

    for image_tool_dir, configs in results.items():
        results[image_tool_dir] = sorted(
            configs,
            key=lambda c: c.timestamp,
            reverse=True,
        )
        if by_time == "latest":
            results[image_tool_dir] = [results[image_tool_dir][0]]

    return list(itertools.chain(*[result for result in results.values() if result]))


def find_one(*args, **kwargs) -> artifact.ScanConfiguration:
    configs = find(*args, **kwargs)

    if not configs:
        raise RuntimeError(f"no results found for {kwargs}")

    if len(configs) > 1:
        raise RuntimeError(f"multiple results found for {kwargs}")

    return configs[0]


def load_by_descriptions(
    descriptions: list[str],
    year_max_limit: int | None = None,
    year_from_cve_only: bool = False,
    skip_sbom_results: bool = False,
    store_root: str | None = None,
) -> list[artifact.ScanResult]:
    results = []
    for description in descriptions:
        config = find_one(by_description=description, store_root=store_root)
        result = load(
            config=config,
            year_max_limit=year_max_limit,
            year_from_cve_only=year_from_cve_only,
            store_root=store_root,
        )
        if skip_sbom_results and result.packages is not None:
            # note: we look at a NONE value, not just an empty list
            logging.debug(f"skipping SBOM result from {description}")
            continue
        results.append(result)
    return results


def load_all(
    configs: list[artifact.ScanConfiguration],
    year_max_limit: int | None = None,
    year_from_cve_only: bool = False,
    store_root: str | None = None,
) -> list[artifact.ScanResult]:
    return [
        load(
            config,
            year_max_limit=year_max_limit,
            year_from_cve_only=year_from_cve_only,
            store_root=store_root,
        )
        for config in configs
    ]


# note: we read the raw data and metadata and recompute the matches
def load(
    config: artifact.ScanConfiguration,
    year_max_limit: int | None = None,
    year_from_cve_only: bool = False,
    store_root: str | None = None,
) -> artifact.ScanResult:
    data_path, metadata_path = store_paths(config, store_root=store_root)
    logging.debug(f"loading result config={config!r} location={data_path!r}")

    with open(data_path, encoding="utf-8") as data_file:
        data_json = data_file.read()

    with open(metadata_path, encoding="utf-8") as metadata_file:
        metadata_dict = json.load(metadata_file)

    tool_name_and_label = metadata_dict["config"]["tool_name"]
    tool_name = tool_name_and_label.split("[")[0]
    selected_tool = tools[tool_name]

    result = selected_tool.parse(data_json, config=config)

    keys = {}
    if issubclass(selected_tool, vulnerability_scanner.VulnerabilityScanner):
        keys["matches"] = result
    elif issubclass(selected_tool, sbom_generator.SBOMGenerator):
        keys["packages"] = result
    else:
        raise RuntimeError("unknown tool type")

    results = {**metadata_dict, **keys}

    result_obj: artifact.ScanResult = artifact.ScanResult.from_dict(results)  # type: ignore[attr-defined]

    # TODO: allow for searching for more results until there is a matching digest
    if config.image_digest and result_obj.config.image_digest != config.image_digest:
        raise RuntimeError(
            f"scan config has a specific image digest={config.image_digest} that differs from the fetched digest={result_obj.config.image_digest}",
        )

    if year_max_limit:
        results_list = filter_by_year(
            [result_obj],
            year_max_limit=int(year_max_limit),
            year_from_cve_only=year_from_cve_only,
        )
        result_obj = results_list[0]

    return result_obj


def list_all_metadata_json(store_root: str | None = None):
    json_path = tool.results_path(store_root=store_root)
    return glob.glob(f"{json_path}/*/*/*/metadata.json")


def list_all_configs(store_root: str | None = None) -> list[artifact.ScanConfiguration]:
    results = []
    for metadata_file in list_all_metadata_json(store_root=store_root):
        with open(metadata_file, encoding="utf-8") as metadata_file:
            metadata_dict = json.load(metadata_file)
            results.append(
                artifact.ScanConfiguration.from_dict(metadata_dict["config"]),  # type: ignore[attr-defined]
            )
    return sorted(results)


# filter_by_year filters out CVE vuln IDs above a given year. We attempt to normalize all vuln IDs to CVEs,
# but will include any if normalization fails.
def filter_by_year(
    results: list[artifact.ScanResult],
    year_max_limit: int,
    year_from_cve_only: bool = False,
) -> list[artifact.ScanResult]:
    results_copy = copy.deepcopy(results)

    for i, r in enumerate(results):
        results_copy[i].matches = []

        if not r.matches:
            continue

        for m in r.matches:
            year = m.vulnerability.effective_year(by_cve=year_from_cve_only)

            if not year or year <= year_max_limit:
                results_copy[i].matches.append(m)  # type: ignore[union-attr]

    return results_copy



import enum
from dataclasses import dataclass

from yardstick import artifact, comparison


class DeltaType(enum.Enum):
    Unknown = "Unknown"
    FixedFalseNegative = "FixedFalseNegative"
    FixedFalsePositive = "FixedFalsePositive"
    NewFalseNegative = "NewFalseNegative"
    NewFalsePositive = "NewFalsePositive"


@dataclass
class Delta:
    tool: str
    package_name: str
    package_version: str
    vulnerability_id: str
    added: bool
    label: str | None = None

    @property
    def is_improved(self) -> bool | None:
        if self.outcome in {DeltaType.FixedFalseNegative, DeltaType.FixedFalsePositive}:
            return True
        if self.outcome in {DeltaType.NewFalseNegative, DeltaType.NewFalsePositive}:
            return False
        return None

    @property
    def commentary(self) -> str:
        commentary = ""
        # if self.is_improved and self.label == artifact.Label.TruePositive.name:
        if self.outcome == DeltaType.FixedFalseNegative:
            commentary = "(this is a new TP 🙌)"
        elif self.outcome == DeltaType.FixedFalsePositive:
            commentary = "(got rid of a former FP 🙌)"
        elif self.outcome == DeltaType.NewFalsePositive:
            commentary = "(this is a new FP 😱)"
        elif self.outcome == DeltaType.NewFalseNegative:
            commentary = "(this is a new FN 😱)"

        return commentary

    @property
    def outcome(self) -> DeltaType:
        # TODO: this would be better handled post init and set I think
        if not self.label:
            return DeltaType.Unknown

        if not self.added:
            # the tool which found the unique result is the reference tool...
            if self.label == artifact.Label.TruePositive.name:
                # drats! we missed a case (this is a new FN)
                return DeltaType.NewFalseNegative
            elif artifact.Label.FalsePositive.name in self.label:
                # we got rid of a FP! ["hip!", "hip!"]
                return DeltaType.FixedFalsePositive
        else:
            # the tool which found the unique result is the current tool...
            if self.label == artifact.Label.TruePositive.name:
                # highest of fives! we found a new TP that the previous tool release missed!
                return DeltaType.FixedFalseNegative
            elif artifact.Label.FalsePositive.name in self.label:
                # welp, our changes resulted in a new FP... not great, maybe not terrible?
                return DeltaType.NewFalsePositive

        return DeltaType.Unknown


def compute_deltas(
    comparisons_by_result_id: dict[str, comparison.AgainstLabels],
    reference_tool: str,
    relative_comparison: comparison.ByPreservedMatch,
):
    deltas = []
    for result in relative_comparison.results:
        label_comparison = comparisons_by_result_id[result.ID]
        for unique_match in relative_comparison.unique[result.ID]:
            labels = label_comparison.labels_by_match[unique_match.ID]
            if not labels:
                label = "(unknown)"
            elif len(set(labels)) > 1:
                label = ", ".join([la.name for la in labels])
            else:
                label = labels[0].name

            delta = Delta(
                tool=result.config.tool,
                package_name=unique_match.package.name,
                package_version=unique_match.package.version,
                vulnerability_id=unique_match.vulnerability.id,
                added=result.config.tool != reference_tool,
                label=label,
            )
            deltas.append(delta)
    return deltas



from asyncio import Future, ensure_future
from typing import List, Optional

# from pygments.styles.monokai import MonokaiStyle as PygmentsStyle
from prompt_toolkit.application import Application
from prompt_toolkit.filters import has_focus
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.layout import (
    Dimension,
    Float,
    FloatContainer,
    FormattedTextControl,
    HSplit,
    Layout,
    VSplit,
    Window,
)
from prompt_toolkit.styles import Style, style_from_pygments_cls
from pygments.styles.native import NativeStyle as PygmentsStyle

from yardstick import artifact
from yardstick.cli.explore.image_labels.display_pane import DisplayPane
from yardstick.cli.explore.image_labels.label_manager import LabelManager
from yardstick.cli.explore.image_labels.label_selection_pane import LabelSelectionPane
from yardstick.cli.explore.image_labels.result_selection_pane import ResultSelectionPane
from yardstick.cli.explore.image_labels.text_dialog import TextDialog


class Controller:
    def __init__(
        self,
        result: artifact.ScanResult,
        label_entries: List[artifact.LabelEntry],
        lineage: List[str],
        filter_spec: str = "",
    ):
        self.result = result
        self.lineage = lineage
        self.label_manager = LabelManager(result, label_entries, lineage)

        self.label_details_pane = LabelSelectionPane(
            self.label_manager,
            dialog_executor=self.show_dialog_as_float,
        )
        self.result_selection_pane = ResultSelectionPane(
            label_manager=self.label_manager,
            dialog_executor=self.show_dialog_as_float,
            match_setter=self.label_details_pane.set_match,
            filter_spec=filter_spec,
        )
        self.result_details_pane = DisplayPane(
            self.result_selection_pane.get_formatted_result_details,
            title="Result Details",
        )
        self.cve_details_pane = DisplayPane(
            self.result_selection_pane.get_formatted_cve_details,
            title="CVE Description",
            height=Dimension(preferred=8),
        )

        self.root_container = self._layout()
        self.application: Optional[Application] = None

    def show_dialog_as_float(self, dialog, done_callback):
        async def coroutine():
            float_ = Float(content=dialog)
            self.root_container.floats.append(float_)

            focused_before = self.application.layout.current_window
            self.application.layout.focus(dialog)
            result = await dialog.future
            self.application.layout.focus(focused_before)

            if float_ in self.root_container.floats:
                self.root_container.floats.remove(float_)

            return result

        ensure_future(coroutine()).add_done_callback(done_callback)

    def _layout(self):
        # note: the root container contains all floats for dialogs
        return FloatContainer(
            content=HSplit(
                [
                    # main header + image + tool info
                    Window(
                        content=FormattedTextControl(self._get_info_text),
                        height=Dimension.exact(3),
                        style="class:status",
                    ),
                    # Main content...
                    VSplit(
                        [
                            # Left side...
                            HSplit(
                                [
                                    self.result_selection_pane,
                                    # TODO: add filter here...
                                ],
                            ),
                            # Right side...
                            HSplit(
                                [
                                    self.cve_details_pane,
                                    self.label_details_pane,
                                    self.result_details_pane,
                                ],
                                width=Dimension(preferred=100),
                            ),
                        ],
                    ),
                    # Statusbar...
                    VSplit(
                        [
                            Window(
                                content=FormattedTextControl(self._get_keybinding_text),
                                # height=Dimension.exact(2),
                                height=Dimension.exact(1),
                                style="class:status",
                            ),
                            Window(
                                content=FormattedTextControl(self._get_status_text),
                                height=Dimension.exact(1),
                                style="class:status",
                                dont_extend_width=True,
                            ),
                        ],
                    ),
                ],
            ),
            floats=[],
        )

    @staticmethod
    def _styles() -> Style:
        pygments_style = dict(
            style_from_pygments_cls(pygments_style_cls=PygmentsStyle).style_rules,
        )

        style_dict = {
            "select-match-box cursor-line": "nounderline bold fg:#fcca03",
            "pane-title cursor-line": "nounderline bold reverse",
            "select-label-box cursor-line": "nounderline bold fg:#fcca03",
            "status": "reverse",
            "status.title": "bold",
            "status.key": "bold",
            "status.history-metric": "bg:#ff0066 bold",
            "filter-title": "fg:#fcca03",
            "filter-toolbar-text": "italic",
            "radio-selected": "bold fg:ansired underline",
            "radio-checked": "bold fg:ansired reverse",
        }

        style_dict.update(**pygments_style)

        return Style.from_dict(style_dict)

    def _keybindings(self):
        bindings = KeyBindings()
        bindings.add("tab")(focus_next)
        bindings.add("s-tab")(focus_previous)

        @bindings.add("c-s")
        def _save(_):
            self.label_manager.write()

        @bindings.add("c-c")
        def _exit(event):
            if self.label_manager.history.total_events():

                def done(fut: Future):
                    if fut.done() and fut.result() is True:
                        self.label_manager.write()
                    event.app.exit()

                self.show_dialog_as_float(
                    dialog=TextDialog(
                        title="Exiting...",
                        label_text="There are unsaved changes.",
                        text_area=False,
                        ok_button_text="Save",
                        cancel_button_text="Don't Save",
                    ),
                    done_callback=done,
                )
            else:
                event.app.exit()

        @bindings.add("c-z")
        def _undo(_):
            self.label_manager.undo()
            self.application.invalidate()

        @bindings.add("c-y")
        def _redo(_):
            self.label_manager.redo()
            self.application.invalidate()

        return bindings

    def _get_info_text(self):
        return [
            ("class:status.title", "Image "),
            ("class:status", self.result.config.image),
            ("class:status", "  "),
            ("class:status.title", "Tool "),
            ("class:status", f"{self.result.config.tool}"),
            ("class:status", "\n"),
            ("class:status.title", "Image Parents "),
            ("class:status", str(self.lineage)),
            ("class:status", "\n"),
            ("class:status.title", "Matches: "),
            (
                "class:status.history-metric",
                f"{len(self.label_manager.match_select_entries)}",
            ),
            ("class:status", "  "),
            ("class:status.title", "Applied Labels: "),
            ("class:status.history-metric", f"{self.label_manager.applied_labels()}"),
            ("class:status", "  "),
            ("class:status.title", "Labeled Matches: "),
            ("class:status.history-metric", f"{self.label_manager.matches_labeled()}"),
            ("class:status", "  "),
            ("class:status.title", "Unlabeled Matches: "),
            (
                "class:status.history-metric",
                f"{self.label_manager.matches_not_labeled()}",
            ),
            ("class:status", "  "),
            ("class:status.title", "F1 Score: "),
            ("class:status.history-metric", f"{self.label_manager.f1_score()}"),
            ("class:status", "  "),
        ]

    def _get_status_text(self):
        results = []

        undoneChanges = self.label_manager.history.undone_events()
        currentChanges = self.label_manager.history.total_events() - undoneChanges

        # fun, but unnecessary
        # if undoneChanges:
        #     results += [
        #         ("class:status", "undone changes: "),
        #         ("class:status.history-metric", f"{undoneChanges}"),
        #         ("class:status", "  "),
        #     ]

        if currentChanges:
            results += [
                ("class:status", "unsaved changes: "),
                ("class:status.history-metric", f"{currentChanges}"),
                ("class:status", "  "),
            ]

        return results

    def _get_keybinding_text(self):
        results = [
            ("class:status.key", " ^C"),
            ("class:status", " exit ▏"),
            ("class:status.key", "^S"),
            ("class:status", " save ▏"),
            ("class:status.key", "^Z/Y"),
            ("class:status", " undo/redo ▏"),
            ("class:status.key", "Tab"),
            ("class:status", " next pane ▏"),
        ]

        results_selected = has_focus(self.result_selection_pane)()
        labels_selected = has_focus(self.label_details_pane)()
        if results_selected:
            results += [
                ("class:status.key", "T/F/?"),
                ("class:status", " mark label ▏"),
                ("class:status.key", "E"),
                ("class:status", " edit note ▏"),
                ("class:status.key", "J"),
                ("class:status", " edit json ▏"),
                ("class:status.key", "DEL"),
                ("class:status", " delete ▏"),
                ("class:status.key", "/"),
                ("class:status", " filter ▏"),
            ]

        if labels_selected:
            results += [
                ("class:status.key", " E"),
                ("class:status", " edit label note ▏"),
                ("class:status.key", " J"),
                ("class:status", " edit label json ▏"),
                ("class:status.key", " DEL"),
                ("class:status", " delete label ▏"),
            ]

        return results

    def run(self):
        self.application: Application = Application(
            layout=Layout(
                self.root_container,
                focused_element=self.result_selection_pane,
            ),
            key_bindings=self._keybindings(),
            enable_page_navigation_bindings=True,
            mouse_support=False,  # trust me, this is better
            style=self._styles(),
            full_screen=True,
        )

        return self.application.run()


def run(
    result: artifact.ScanResult,
    label_entries: List[artifact.LabelEntry],
    lineage: List[str],
    filter_spec: str = "",
):
    Controller(result, label_entries, lineage, filter_spec).run()



import typing

from yardstick.artifact import Label, Package
from yardstick.validate.delta import Delta, DeltaType, compute_deltas

import pytest
from unittest.mock import MagicMock
from yardstick.comparison import AgainstLabels, ByPreservedMatch


@pytest.mark.parametrize(
    "tool, package_name, package_version, vulnerability_id, added, label, expected_outcome, expected_is_improved, expected_commentary",
    [
        (
            "scanner1",
            "libc",
            "2.29",
            "CVE-2023-1234",
            True,
            Label.TruePositive.name,
            DeltaType.FixedFalseNegative,
            True,
            "(this is a new TP 🙌)",
        ),
        (
            "scanner1",
            "nginx",
            "1.17",
            "CVE-2023-0002",
            False,
            Label.FalsePositive.name,
            DeltaType.FixedFalsePositive,
            True,
            "(got rid of a former FP 🙌)",
        ),
        (
            "scanner2",
            "bash",
            "5.0",
            "CVE-2023-5678",
            False,
            Label.TruePositive.name,
            DeltaType.NewFalseNegative,
            False,
            "(this is a new FN 😱)",
        ),
        (
            "scanner3",
            "zlib",
            "1.2.11",
            "CVE-2023-8888",
            True,
            Label.FalsePositive.name,
            DeltaType.NewFalsePositive,
            False,
            "(this is a new FP 😱)",
        ),
        (
            "scanner4",
            "openssl",
            "1.1.1",
            "CVE-2023-0001",
            True,
            None,
            DeltaType.Unknown,
            None,
            "",
        ),
    ],
)
def test_delta_properties(
    tool,
    package_name,
    package_version,
    vulnerability_id,
    added,
    label,
    expected_outcome,
    expected_is_improved,
    expected_commentary,
):
    """Test Delta properties is_improved, outcome, and commentary based on logical combinations."""

    delta = Delta(
        tool=tool,
        package_name=package_name,
        package_version=package_version,
        vulnerability_id=vulnerability_id,
        added=added,
        label=label,
    )

    assert delta.outcome == expected_outcome
    assert delta.is_improved == expected_is_improved
    assert delta.commentary == expected_commentary


@pytest.fixture
def reference_result():
    """Fixture for creating a mock reference result."""
    return MagicMock(name="reference_results", ID="reference", config=MagicMock(tool="reference"))


@pytest.fixture
def candidate_result():
    """Fixture for creating a mock candidate result."""
    return MagicMock(name="candidate_results", ID="candidate", config=MagicMock(tool="candidate"))


@pytest.fixture
def comparisons_by_result_id():
    """Fixture for setting up comparisons with expected label data (source of truth)."""
    comparison = {
        # skip post init calculations on against labels, since
        # we're setting the comparison results directly below
        "reference": typing.cast(AgainstLabels, object.__new__(AgainstLabels)),
        "candidate": typing.cast(AgainstLabels, object.__new__(AgainstLabels)),
    }
    comparison["reference"].labels_by_match = {
        "match1": [Label.TruePositive],
        "match2": [Label.TruePositive],
        "match3": [Label.FalsePositive],
        "match4": [Label.FalsePositive],
    }
    comparison["candidate"].labels_by_match = {
        "match1": [Label.TruePositive],
        "match2": [Label.TruePositive],
        "match3": [Label.FalsePositive],
        "match4": [Label.FalsePositive],
    }
    return comparison


@pytest.fixture
def relative_comparison(reference_result, candidate_result):
    """Fixture for creating a mock relative comparison of reference and candidate."""
    match1 = MagicMock(
        name="match1",
        ID="match1",
        package=Package(name="libc", version="2.29"),
        vulnerability=MagicMock(id="CVE-2023-1234"),
    )
    match2 = MagicMock(
        name="match2",
        ID="match2",
        package=Package(name="nginx", version="1.17"),
        vulnerability=MagicMock(id="CVE-2023-0002"),
    )
    match3 = MagicMock(
        name="match3",
        ID="match3",
        package=Package(name="openssl", version="1.1.1"),
        vulnerability=MagicMock(id="CVE-2023-5678"),
    )
    match4 = MagicMock(
        name="match4",
        ID="match4",
        package=Package(name="zlib", version="1.2.11"),
        vulnerability=MagicMock(id="CVE-2023-8888"),
    )

    result = ByPreservedMatch(
        results=[reference_result, candidate_result],
    )
    result.unique = {
        "reference": [match2, match3],
        "candidate": [match1, match4],
    }
    return result


def test_compute_deltas(comparisons_by_result_id, relative_comparison):
    """Test compute_deltas with realistic comparisons between reference and candidate results."""
    deltas = compute_deltas(
        comparisons_by_result_id=comparisons_by_result_id,
        reference_tool="reference",
        relative_comparison=relative_comparison,
    )

    expected_deltas = [
        Delta(
            tool="reference",
            package_name="nginx",
            package_version="1.17",
            vulnerability_id="CVE-2023-0002",
            added=False,
            label="TruePositive",
        ),
        Delta(
            tool="reference",
            package_name="openssl",
            package_version="1.1.1",
            vulnerability_id="CVE-2023-5678",
            added=False,
            label="FalsePositive",
        ),
        Delta(
            tool="candidate",
            package_name="libc",
            package_version="2.29",
            vulnerability_id="CVE-2023-1234",
            added=True,
            label="TruePositive",
        ),
        Delta(
            tool="candidate",
            package_name="zlib",
            package_version="1.2.11",
            vulnerability_id="CVE-2023-8888",
            added=True,
            label="FalsePositive",
        ),
    ]

    assert len(deltas) == len(expected_deltas)
    for idx, actual in enumerate(deltas):
        assert actual == expected_deltas[idx], f"unequal at {idx}"



import pytest
import tarfile
import os
import json
from tempfile import TemporaryDirectory
from unittest import mock

import zstandard as zstd
import xxhash

from yardstick.tool.grype import (
    Grype,
    GrypeProfile,
    handle_legacy_archive,
    handle_zstd_archive,
)


def test_grype_profiles():
    profile_arg = {"name": "test-profile", "config_path": "test-config-path"}
    profile = GrypeProfile(**profile_arg)
    with mock.patch("subprocess.check_output") as check_output:
        check_output.return_value = bytes("test-output", "utf-8")
        tool = Grype(path="test-path", profile=profile, db_identity="oss")
        tool.capture(image="test-image", tool_input=None)
        assert check_output.call_args.args[0] == [
            "test-path/grype",
            "-o",
            "json",
            "test-image",
            "-c",
            "test-config-path",
        ]


def test_grype_no_profile():
    with mock.patch("subprocess.check_output") as check_output:
        check_output.return_value = bytes("test-output", "utf-8")
        tool = Grype(path="test-path", db_identity="oss")
        tool.capture(image="test-image", tool_input=None)
        assert check_output.call_args.args[0] == [
            "test-path/grype",
            "-o",
            "json",
            "test-image",
        ]


def test_install_from_path():
    with (
        mock.patch("subprocess.check_call") as check_call,
        mock.patch(
            "git.Repo",
        ) as repo,
        mock.patch("os.path.exists") as exists,
        mock.patch(
            "os.makedirs",
        ),
        mock.patch(
            "os.chmod",
        ),
    ):
        check_call.return_value = bytes("test-output", "utf-8")
        exists.return_value = True
        fake_repo = mock.Mock()
        fake_repo.git = mock.Mock()
        fake_repo.untracked_files = []
        git_describe_val = "v0.65.1-1-g74a7a67-dirty"
        hash_of_git_diff = "a29864cf5600b481056b6fa30a21cdbabc15287d"[:8]
        fake_repo.git.describe.return_value = git_describe_val
        fake_repo.git.diff.return_value = "test-diff"  # hash is 'a29864cf5600b481056b6fa30a21cdbabc15287d'
        repo.return_value = fake_repo
        version_str = "path:/where/grype/is/cloned"
        normalized_version_str = version_str.replace("/", "_").removeprefix("path:")
        expected_grype_path = f".yardstick/tools/grype/{normalized_version_str}/{git_describe_val}-{hash_of_git_diff}/local_install"
        tool = Grype.install(
            version=version_str,
            path=".yardstick/tools/grype/path:_where_grype_is_cloned",
            update_db=False,
        )
        assert tool.path == expected_grype_path


def create_legacy_archive_with_metadata(archive_path, metadata):
    with tarfile.open(archive_path, "w:gz") as tar:
        metadata_path = os.path.join(os.path.dirname(archive_path), "metadata.json")
        with open(metadata_path, "w") as f:
            json.dump(metadata, f)
        tar.add(metadata_path, arcname="metadata.json")
        os.remove(metadata_path)


def create_zstd_archive_with_db(archive_path, db_content):
    with TemporaryDirectory() as temp_dir:
        db_path = os.path.join(temp_dir, "vulnerability.db")
        with open(db_path, "wb") as f:
            f.write(db_content)

        tar_path = os.path.join(temp_dir, "archive.tar")
        with tarfile.open(tar_path, "w") as tar:
            tar.add(db_path, arcname="vulnerability.db")

        with open(tar_path, "rb") as tar_file, open(archive_path, "wb") as zstd_file:
            dctx = zstd.ZstdCompressor()
            dctx.copy_stream(tar_file, zstd_file)


@pytest.fixture
def legacy_archive(tmp_path):
    archive_path = tmp_path / "db.tar.gz"
    metadata = {"checksum": "12345"}
    create_legacy_archive_with_metadata(archive_path, metadata)
    return archive_path, metadata["checksum"]


@pytest.fixture
def zstd_archive(tmp_path):
    archive_path = tmp_path / "db.tar.zst"
    db_content = b"dummy database content"
    hasher = xxhash.xxh64()
    hasher.update(db_content)
    expected_checksum = hasher.hexdigest()
    create_zstd_archive_with_db(archive_path, db_content)
    return archive_path, expected_checksum


def test_handle_legacy_archive(legacy_archive):
    archive_path, expected_checksum = legacy_archive
    assert handle_legacy_archive(str(archive_path)) == expected_checksum


def test_handle_zstd_archive(zstd_archive):
    archive_path, expected_checksum = zstd_archive
    assert handle_zstd_archive(str(archive_path)) == expected_checksum



from dataclasses import dataclass, field, InitVar
from typing import Optional

from yardstick import comparison
from yardstick.validate.delta import Delta


@dataclass
class GateConfig:
    max_f1_regression: float = 0.0
    max_new_false_negatives: int = 0
    max_unlabeled_percent: int = 0
    max_year: int | None = None
    reference_tool_label: str = "reference"
    candidate_tool_label: str = "candidate"
    # only consider matches from these namespaces when judging results
    allowed_namespaces: list[str] = field(default_factory=list)
    # fail this gate unless all of these namespaces are present
    required_namespaces: list[str] = field(default_factory=list)
    fail_on_empty_match_set: bool = True


@dataclass
class GateInputResultConfig:
    id: str
    tool: str
    tool_label: str


@dataclass
class GateInputDescription:
    image: str
    configs: list[GateInputResultConfig] = field(default_factory=list)


@dataclass
class Gate:
    reference_comparison: InitVar[Optional[comparison.LabelComparisonSummary]]
    candidate_comparison: InitVar[Optional[comparison.LabelComparisonSummary]]

    config: GateConfig

    input_description: GateInputDescription
    reasons: list[str] = field(default_factory=list)
    deltas: list[Delta] = field(default_factory=list)

    def __post_init__(
        self,
        reference_comparison: Optional[comparison.LabelComparisonSummary],
        candidate_comparison: Optional[comparison.LabelComparisonSummary],
    ):
        if not reference_comparison or not candidate_comparison:
            return

        reasons = []

        reference_f1_score = reference_comparison.f1_score
        current_f1_score = candidate_comparison.f1_score
        if current_f1_score < reference_f1_score - self.config.max_f1_regression:
            reasons.append(
                f"current F1 score is lower than the latest release F1 score: candidate_score={current_f1_score:0.2f} reference_score={reference_f1_score:0.2f} image={self.input_description.image}"
            )

        if candidate_comparison.indeterminate_percent > self.config.max_unlabeled_percent:
            reasons.append(
                f"current indeterminate matches % is greater than {self.config.max_unlabeled_percent}%: candidate={candidate_comparison.indeterminate_percent:0.2f}% image={self.input_description.image}"
            )

        reference_fns = reference_comparison.false_negatives
        candidate_fns = candidate_comparison.false_negatives
        if candidate_fns > reference_fns + self.config.max_new_false_negatives:
            reasons.append(
                f"current false negatives is greater than the latest release false negatives: candidate={candidate_fns} reference={reference_fns} image={self.input_description.image}"
            )

        self.reasons = reasons

    def passed(self) -> bool:
        return len(self.reasons) == 0

    @classmethod
    def failing(cls, reasons: list[str], input_description: GateInputDescription):
        """failing bypasses Gate's normal validation calculating and returns a
        gate that is failing for the reasons given."""
        return cls(
            reference_comparison=None,
            candidate_comparison=None,
            config=GateConfig(),
            reasons=reasons,
            input_description=input_description,
        )

    @classmethod
    def passing(cls, input_description: GateInputDescription):
        """passing bypasses a Gate's normal validation and returns a gate that is passing."""
        return cls(
            reference_comparison=None,
            candidate_comparison=None,
            config=GateConfig(),
            reasons=[],  # a gate with no reason to fail is considered passing
            input_description=input_description,
        )



import pytest

from yardstick import utils


@pytest.mark.parametrize(
    ("input", "expected_year"),
    [
        ("CVE-2016-2781", 2016),
        ("CVE-1989-18276", None),
        ("CVE-20222-18276", None),
        ("ALAS-2019-1234", 2019),
        ("ALASRUBY2.6-2023-006", 2023),
        ("ALASSELINUX-NG-2023-001", 2023),
        ("ALASKERNEL-5.4-2023-043", 2023),
        ("ELSA-2023-6162", 2023),
    ],
)
def test_parse_year_from_id(input, expected_year):
    assert utils.parse_year_from_id(input) == expected_year



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import Any, Sequence

import mergedeep  # type: ignore[import]
import re
import yaml
from dataclass_wizard import asdict, fromdict  # type: ignore[import]

from yardstick import artifact, validate
from yardstick.store import config as store_config

DEFAULT_CONFIGS = (
    ".yardstick.yaml",
    "yardstick.yaml",
)


@dataclass()
class Profiles:
    data: dict[str, dict[str, str]] = field(default_factory=dict)

    def get(self, tool_name: str, profile: str):
        return self.data.get(tool_name, {}).get(profile, {})


@dataclass()
class Tool:
    name: str
    version: str
    label: str | None = None
    produces: str | None = None
    takes: str | None = None
    profile: str | None = None
    refresh: bool = True

    @property
    def short(self):
        return f"{self.name}@{self.version}"


@dataclass()
class ScanMatrix:
    images: list[str] = field(default_factory=list)
    tools: list[Tool] = field(default_factory=list)

    DIGEST_REGEX = re.compile(r"(?P<digest>sha256:[a-fA-F0-9]{64})")

    def __post_init__(self):
        for idx, tool in enumerate(self.tools):
            (
                self.tools[idx].name,
                self.tools[idx].version,
            ) = artifact.ScanRequest.render_tool(tool.short).split("@", 1)

        # flatten elements in images (in case yaml anchores are used)
        images = []
        for image in self.images:
            if isinstance(image, list):
                images += image
            if image.startswith("["):
                # technically yaml anchors to lists of lists are interpreted as strings... which is terrible
                images += yaml.safe_load(image)
            else:
                images += [image]
        self.images = images
        invalid = [image for image in images if not ScanMatrix.is_valid_oci_reference(image)]
        if invalid:
            raise ValueError(f"all images must be complete OCI references, but {' '.join(invalid)} are not")

    @staticmethod
    def is_valid_oci_reference(image: str) -> bool:
        host, _, repository, _, digest = ScanMatrix.parse_oci_reference(image)
        return all([host, repository, digest]) and bool(ScanMatrix.DIGEST_REGEX.match(digest or "")) and ("." in host or "localhost" in host)

    @staticmethod
    def parse_oci_reference(image: str) -> tuple[str, str, str, str, str]:
        host = ""
        path = ""
        host_and_path = ""
        repository = ""
        tag = ""
        digest = ""

        if "@" in image:
            pre_digest, digest = image.rsplit("@", 1)
        else:
            pre_digest = image

        if ":" in pre_digest:
            pre_tag, tag = pre_digest.rsplit(":", 1)
        else:
            pre_tag = pre_digest

        if "/" in pre_tag:
            host_and_path, repository = pre_tag.rsplit("/", 1)
        else:
            repository = pre_tag

        if host_and_path:
            parts = host_and_path.split("/")
            host = parts[0]
            path = "/".join(parts[1:]) if len(parts) > 1 else ""

        return host, path, repository, tag, digest


@dataclass()
class Validation(validate.GateConfig):
    name: str = "default"


@dataclass()
class ResultSet:
    description: str = ""
    declared: list[artifact.ScanRequest] = field(default_factory=list)
    matrix: ScanMatrix = field(default_factory=ScanMatrix)
    validations: list[Validation] = field(default_factory=list)

    def images(self) -> list[str]:
        return self.matrix.images + [req.image for req in self.declared]

    def scan_requests(self) -> list[artifact.ScanRequest]:
        rendered = []
        for image in self.matrix.images:
            for tool in self.matrix.tools:
                rendered.append(
                    artifact.ScanRequest(
                        image=image,
                        tool=tool.short,
                        label=tool.label,
                        profile=tool.profile,
                        provides=tool.produces,
                        takes=tool.takes,
                        refresh=tool.refresh,
                    ),
                )
        return self.declared + rendered


@dataclass()
class Application:
    store_root: str = store_config.DEFAULT_STORE_ROOT
    profile_path: str = ".yardstick.profiles.yaml"
    profiles: Profiles = field(default_factory=Profiles)
    result_sets: dict[str, ResultSet] = field(default_factory=dict)
    default_max_year: int | None = None
    derive_year_from_cve_only: bool = False

    def max_year_for_any_result_set(self, result_sets: list[str]) -> int | None:
        years = []
        for result_set in result_sets:
            m = self.max_year_for_result_set(result_set)
            if m is not None:
                years.append(m)

        if not years:
            return None

        return max(years)

    def max_year_for_result_set(self, result_set: str) -> int | None:
        """return the max year needed by any validation on the result set, or default_max_year"""
        rs = self.result_sets.get(result_set, None)
        years = []
        if rs is not None:
            for gate in rs.validations:
                if gate.max_year is not None:
                    years.append(gate.max_year)
                elif self.default_max_year is not None:
                    years.append(self.default_max_year)

        if years:
            return max(years)

        return self.default_max_year


def clean_dict_keys(d):
    new = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = clean_dict_keys(v)
        new[k.replace("-", "_")] = v
    return new


def yaml_decoder(data) -> dict[Any, Any]:
    return clean_dict_keys(yaml.safe_load(data))


def load(
    path: str | Sequence[str] = DEFAULT_CONFIGS,
) -> Application:
    cfg = _load_paths(path)

    if not cfg:
        msg = "no config found"
        raise RuntimeError(msg)

    if cfg.profile_path:
        try:
            with open(cfg.profile_path, encoding="utf-8") as yaml_file:
                profile = Profiles(yaml_decoder(yaml_file))
        except FileNotFoundError:
            profile = Profiles({})
        cfg.profiles = profile

    return cfg


def _load_paths(
    path: str | Sequence[str],
) -> Application | None:
    if not path:
        path = DEFAULT_CONFIGS

    if isinstance(path, str):
        if path == "":
            path = DEFAULT_CONFIGS
        else:
            return _load(path)

    if isinstance(path, (list, tuple)):
        for p in path:
            if not os.path.exists(p):
                continue

            return _load(p)

        # use the default application config
        return Application()

    msg = f"invalid path type {type(path)}"
    raise ValueError(msg)


def _load(path: str) -> Application:
    with open(path, encoding="utf-8") as f:
        app_object = yaml.load(f.read(), yaml.SafeLoader) or {}  # noqa: S506 (since our loader is using the safe loader)
        # we need a full default application config first then merge the loaded config on top.
        # Why? dataclass_wizard.fromdict() will create instances from the dataclass default
        # and NOT the field definition from the container. So it is possible to specify a
        # single field in the config and all other fields would be set to the default value
        # based on the dataclass definition and not any field(default_factory=...) hints
        # from the containing class.
        instance = asdict(Application())

        mergedeep.merge(instance, app_object)
        cfg = fromdict(
            Application,
            instance,
        )

    if cfg is None:
        msg = "parsed empty config"
        raise FileNotFoundError(msg)

    return cfg



import dataclasses
import enum
import logging
from typing import Any

import click
import importlib_metadata
import yaml

from yardstick import store
from yardstick.cli import config, label, result, validate


@click.option("--verbose", "-v", default=False, help="show logs", is_flag=True)
@click.option("--config", "-c", "config_path", default="", help="override config path")
@click.group(
    help="Tool for parsing and comparing the vulnerability report output from multiple tools.",
)
@click.pass_context
def cli(ctx, verbose: bool, config_path: str):
    import logging.config

    # initialize yardstick based on the current configuration and
    # set the config object to click context to pass to subcommands
    ctx.obj = config.load(config_path)
    store.config.set_values(store_root=ctx.obj.store_root)

    log_level = "INFO"
    if verbose:
        log_level = "DEBUG"

    logging.config.dictConfig(
        {
            "version": 1,
            "formatters": {
                "standard": {
                    "format": "%(asctime)s [%(levelname)s] %(message)s",
                    "datefmt": "",
                },
            },
            "handlers": {
                "default": {
                    "level": log_level,
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stderr",
                },
            },
            "loggers": {
                "": {  # root logger
                    "handlers": ["default"],
                    "level": log_level,
                },
            },
        },
    )


@cli.command(name="config", help="show the application config")
@click.pass_obj
def show_config(cfg: config.Application):
    logging.info("showing application config")

    class IndentDumper(yaml.Dumper):
        def increase_indent(
            self,
            flow: bool = False,
            indentless: bool = False,
        ) -> None:
            return super().increase_indent(flow, False)

    def enum_asdict_factory(data: list[tuple[str, Any]]) -> dict[Any, Any]:
        # prevents showing oddities such as
        #
        #   wolfi:
        #       request_timeout: 125
        #       runtime:
        #       existing_input: !!python/object/apply:vunnel.provider.InputStatePolicy
        #           - keep
        #       existing_results: !!python/object/apply:vunnel.provider.ResultStatePolicy
        #           - delete-before-write
        #       on_error:
        #           action: !!python/object/apply:vunnel.provider.OnErrorAction
        #           - fail
        #           input: !!python/object/apply:vunnel.provider.InputStatePolicy
        #           - keep
        #           results: !!python/object/apply:vunnel.provider.ResultStatePolicy
        #           - keep
        #           retry_count: 3
        #           retry_delay: 5
        #       result_store: !!python/object/apply:vunnel.result.StoreStrategy
        #           - flat-file
        #
        # and instead preferring:
        #
        #   wolfi:
        #       request_timeout: 125
        #       runtime:
        #       existing_input: keep
        #       existing_results: delete-before-write
        #       on_error:
        #           action: fail
        #           input: keep
        #           results: keep
        #           retry_count: 3
        #           retry_delay: 5
        #       result_store: flat-file

        def convert_value(obj: Any) -> Any:
            if isinstance(obj, enum.Enum):
                return obj.value
            return obj

        return {k: convert_value(v) for k, v in data}

    cfg_dict = dataclasses.asdict(cfg, dict_factory=enum_asdict_factory)
    print(yaml.dump(cfg_dict, Dumper=IndentDumper, default_flow_style=False))


@cli.command(name="version", help="show the installed version of yardstick")
@click.pass_obj
def version(_: config.Application):
    d = importlib_metadata.distribution("yardstick")
    if not d:
        raise RuntimeError("yardstick install information could not be found")
    print(f"{d.name} {d.version} ({d.locate_file(d.name).parent})")


cli.add_command(validate.validate)
cli.add_command(result.group)
cli.add_command(label.group)



from unittest.mock import patch

import pytest
from yardstick import artifact as art
from yardstick import store


class TestFilterByYear:
    @pytest.fixture()
    def matches(self) -> list[art.Match]:
        matches: list[art.Match] = []
        pkg = art.Package("busybox", "1.34.1-r5")
        for i in range(3):
            vuln = art.Vulnerability(f"CVE-200{i}-1")
            matches.append(art.Match(vulnerability=vuln, package=pkg))

        matches.append(
            art.Match(
                vulnerability=art.Vulnerability("GHSA-52rh-5rpj-c3w6"),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "GHSA-52rh-5rpj-abc7",
                    cve_id="CVE-2000-1234567",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(vulnerability=art.Vulnerability("ELSA-1998-0098"), package=pkg),
        )
        matches.append(
            art.Match(vulnerability=art.Vulnerability("ELSA-2023-0003"), package=pkg),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ELSA-2022-0002",
                    cve_id="CVE-2022-2222",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ELSA-2021-0001",
                    cve_id="CVE-2000-1234567",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ELSA-1999-1234",
                    cve_id="CVE-2021-1234567",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(vulnerability=art.Vulnerability("ALAS-1998-0098"), package=pkg),
        )
        matches.append(
            art.Match(vulnerability=art.Vulnerability("ALAS-2023-0003"), package=pkg),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ALAS-2022-0002",
                    cve_id="CVE-2022-2222",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ALAS-2021-0001",
                    cve_id="CVE-2000-1234567",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ALAS-1999-1234",
                    cve_id="CVE-2021-1234567",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability("ALASKERNEL-1998-0098"),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability("ALASKERNEL-5.1-2023-0003"),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ALASKERNEL-5.1-2022-0002",
                    cve_id="CVE-2022-2222",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ALASKERNEL-5.1-2021-0001",
                    cve_id="CVE-2000-1234567",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ALASKERNEL-5.1-1999-1234",
                    cve_id="CVE-2021-1234567",
                ),
                package=pkg,
            ),
        )
        matches.append(
            art.Match(
                vulnerability=art.Vulnerability(
                    "ALASKERNEL-1999-1234",
                    cve_id="CVE-2021-1234567",
                ),
                package=pkg,
            ),
        )
        return matches

    @pytest.fixture()
    def results(self, matches) -> list[art.ScanResult]:
        results: list[art.ScanResults] = []

        for i in range(1):
            cfg = art.ScanConfiguration(
                image_repo="ubuntu",
                image_digest="123456",
                tool_name="grype",
                tool_version=f"v{i}.0",
            )
            results.append(art.ScanResult(cfg, matches=matches))
        return results

    @pytest.mark.parametrize(
        ("expected", "year_limit", "year_from_cve_only"),
        [
            # below cases can only look at the CVE for year indications
            (
                [
                    # CVE below year limit
                    "CVE-2000-1",
                    "CVE-2001-1",
                    "CVE-2002-1",
                    "GHSA-52rh-5rpj-c3w6",
                    "GHSA-52rh-5rpj-abc7",
                    # above the year limit, but under this configuration we should only look at the CVE for this determination
                    # (which in this case is missing)
                    "ALASKERNEL-5.1-2023-0003",  # note: no cve
                    "ALAS-2023-0003",  # note: no cve
                    "ELSA-2023-0003",  # note: no cve
                    # no CVE found
                    "ELSA-1998-0098",  # note: no cve
                    "ALAS-1998-0098",  # note: no cve
                    "ALASKERNEL-1998-0098",  # note: no cve
                    # CVE below year limit
                    "ELSA-2021-0001",  # note: cve 2000
                    "ALAS-2021-0001",  # note: cve 2000
                    "ALASKERNEL-5.1-2021-0001",  # note: cve 2000
                    # CVE above year limit
                    # "ELSA-1999-1234", # note: cve 2021
                    # "ALAS-1999-1234", # note: cve 2021
                    # "ALASKERNEL-5.1-1999-1234", # note: cve 2021
                    # "ALASKERNEL-1999-1234", # note: cve 2021
                ],
                2002,
                True,
            ),
            (
                [
                    # CVE below year limit
                    "CVE-2000-1",
                    "GHSA-52rh-5rpj-c3w6",
                    "GHSA-52rh-5rpj-abc7",
                    # above the year limit, but under this configuration we should only look at the CVE for this determination
                    # (which in this case is missing)
                    "ALASKERNEL-5.1-2023-0003",  # note: no cve
                    "ALAS-2023-0003",  # note: no cve
                    "ELSA-2023-0003",  # note: no cve
                    # no CVE found
                    "ELSA-1998-0098",  # note: no cve
                    "ALAS-1998-0098",  # note: no cve
                    "ALASKERNEL-1998-0098",  # note: no cve
                    # CVE below year limit
                    "ELSA-2021-0001",  # note: cve 2000
                    "ALAS-2021-0001",  # note: cve 2000
                    "ALASKERNEL-5.1-2021-0001",  # note: cve 2000
                    # CVE above year limit
                    # "CVE-2001-1",
                    # "CVE-2002-1",
                    # "ELSA-1999-1234", # note: cve 2021
                    # "ALAS-1999-1234", # note: cve 2021
                    # "ALASKERNEL-5.1-1999-1234", # note: cve 2021
                    # "ALASKERNEL-1999-1234", # note: cve 2021
                ],
                2000,
                True,
            ),
            (
                [
                    # CVE below year limit
                    "GHSA-52rh-5rpj-c3w6",
                    # above the year limit, but under this configuration we should only look at the CVE for this determination
                    # (which in this case is missing)
                    "ALASKERNEL-5.1-2023-0003",  # note: no cve
                    "ALAS-2023-0003",  # note: no cve
                    "ELSA-2023-0003",  # note: no cve
                    # no CVE found
                    "ELSA-1998-0098",  # note: no cve
                    "ALAS-1998-0098",  # note: no cve
                    "ALASKERNEL-1998-0098",  # note: no cve
                    # CVE above year limit
                    # "CVE-2000-1",
                    # "CVE-2001-1",
                    # "CVE-2002-1",
                    # "ELSA-1999-1234",  # note: cve 2021
                    # "ALAS-1999-1234",  # note: cve 2021
                    # "ALASKERNEL-5.1-1999-1234", # note: cve 2021
                    # "ALASKERNEL-1999-1234", # note: cve 2021
                ],
                1999,
                True,
            ),
            # below cases can look at the primary ID for year indication
            (
                [
                    # ID below year limit
                    "CVE-2000-1",
                    "CVE-2001-1",
                    "CVE-2002-1",
                    "GHSA-52rh-5rpj-c3w6",
                    "GHSA-52rh-5rpj-abc7",
                    # ID below year limit
                    "ELSA-1999-1234",  # note: cve 2021
                    "ALAS-1999-1234",  # note: cve 2021
                    "ALASKERNEL-5.1-1999-1234",  # note: cve 2021
                    "ALASKERNEL-1999-1234",  # note: cve 2021
                    "ELSA-1998-0098",  # note: no cve
                    "ALAS-1998-0098",  # note: no cve
                    "ALASKERNEL-1998-0098",  # note: cve 2021
                    # ID above year limit
                    # "ELSA-2021-0001",  # note: cve 2000
                    # "ALAS-2021-0001",  # note: cve 2000
                    # "ALASKERNEL-5.1-2021-0001",  # note: cve 2000
                    # "ALASKERNEL-5.1-2023-0003",  # note: no cve
                    # "ALAS-2023-0003",  # note: no cve
                    # "ELSA-2023-0003",  # note: no cve
                ],
                2002,
                False,
            ),
            (
                [
                    # ID below year limit
                    "CVE-2000-1",
                    "GHSA-52rh-5rpj-c3w6",
                    "GHSA-52rh-5rpj-abc7",
                    # ID below year limit
                    "ELSA-1999-1234",  # note: cve 2021
                    "ALAS-1999-1234",  # note: cve 2021
                    "ALASKERNEL-5.1-1999-1234",  # note: cve 2021
                    "ALASKERNEL-1999-1234",  # note: cve 2021
                    "ELSA-1998-0098",  # note: no cve
                    "ALAS-1998-0098",  # note: no cve
                    # Invalid ID, but since we don't know how to get the year it's included
                    "ALASKERNEL-1998-0098",  # note: no cve
                    # ID above year limit
                    # "CVE-2001-1",
                    # "CVE-2002-1",
                    # "ALASKERNEL-5.1-2023-0003",  # note: no cve
                    # "ALAS-2023-0003",  # note: no cve
                    # "ELSA-2023-0003",  # note: no cve
                    # "ELSA-2021-0001",  # note: cve 2000
                    # "ALAS-2021-0001",  # note: cve 2000
                    # "ALASKERNEL-5.1-2021-0001",  # note: cve 2000
                ],
                2000,
                False,
            ),
            (
                [
                    # ID below year limit
                    "GHSA-52rh-5rpj-c3w6",
                    "ELSA-1998-0098",  # note: no cve
                    "ALAS-1998-0098",  # note: no cve
                    "ELSA-1999-1234",  # note: cve 2021
                    "ALAS-1999-1234",  # note: cve 2021
                    "ALASKERNEL-5.1-1999-1234",  # note: cve 2021
                    "ALASKERNEL-1999-1234",  # note: cve 2021
                    "ALASKERNEL-1998-0098",  # note: no cve
                    # ID above year limit
                    # "CVE-2000-1",
                    # "CVE-2001-1",
                    # "CVE-2002-1",
                    # "ALASKERNEL-5.1-2023-0003",  # note: no cve
                    # "ALAS-2023-0003",  # note: no cve
                    # "ELSA-2023-0003",  # note: no cve
                ],
                1999,
                False,
            ),
        ],
    )
    def test_filter_by_year(self, expected, year_limit, results, year_from_cve_only):
        assert len(results) == 1

        with patch("yardstick.utils.grype_db.normalize_to_cve", lambda _: None):
            filtered = store.scan_result.filter_by_year(
                results,
                year_limit,
                year_from_cve_only=year_from_cve_only,
            )

        for r in filtered:
            ids = [m.vulnerability.id for m in r.matches]
            assert set(expected) == set(ids)
            assert len(expected) == len(ids)  # duplicate matches must be honored



from asyncio import Future

from prompt_toolkit.application.current import get_app
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import Dimension
from prompt_toolkit.layout.containers import HSplit
from prompt_toolkit.widgets import Button, Dialog, Label, TextArea


class TextDialog:
    def __init__(  # noqa: PLR0913
        self,
        title="",
        label_text="",
        completer=None,
        text_area=True,
        default_value="",
        ok_button_text="OK",
        cancel_button_text="Cancel",
    ):
        self.future = Future()

        def accept_text(buf):
            get_app().layout.focus(ok_button)
            buf.complete_state = None
            return True

        def accept():
            if text_area:
                self.future.set_result(self.text_area.text)
            else:
                self.future.set_result(True)

        kb = KeyBindings()

        kb.add("escape")

        def cancel():
            self.future.set_result(None)

        if text_area:
            self.text_area = TextArea(
                text=default_value or "",
                completer=completer,
                multiline=True,
                width=Dimension(preferred=40),
                height=Dimension(preferred=5),
                accept_handler=accept_text,
            )

        ok_button = Button(
            text=ok_button_text,
            left_symbol="[",
            right_symbol="]",
            handler=accept,
        )
        cancel_button = Button(
            text=cancel_button_text,
            left_symbol="[",
            right_symbol="]",
            handler=cancel,
        )

        layout_items = [Label(text=label_text)]
        if text_area:
            layout_items.append(self.text_area)

        self.dialog = Dialog(
            title=title,
            body=HSplit(layout_items, key_bindings=kb),
            buttons=[ok_button, cancel_button],
            width=Dimension(preferred=80),
            modal=True,
        )

    def __pt_container__(self):
        return self.dialog



import collections
from typing import Dict, List, Set

from yardstick import artifact


def packages_by_vulnerability(
    matches: List[artifact.Match],
) -> Dict[artifact.Vulnerability, Set[artifact.Package]]:
    result: Dict[artifact.Vulnerability, Set[artifact.Package]] = {m.vulnerability: set() for m in matches}
    for match in matches:
        result[match.vulnerability].add(match.package)
    return result


def scan_results_by_image(
    results: list[artifact.ScanResult],
) -> Dict[str, list[artifact.ScanResult]]:
    results_by_image = collections.defaultdict(list)
    for r in results:
        results_by_image[r.config.image].append(r)
    return results_by_image



# Sample images
from unittest.mock import patch, MagicMock

import pytest

from yardstick import comparison
from yardstick.artifact import (
    ScanResult,
    ScanConfiguration,
    Package,
    Vulnerability,
    LabelEntry,
    Label,
    Match,
)
from yardstick.validate import validate_image, GateConfig, Delta


@pytest.fixture()
def compare_results_no_matches():
    return MagicMock(results=[MagicMock(matches=[]), MagicMock(matches=[])])


@pytest.fixture()
def compare_results_identical_matches():
    return MagicMock(
        results=[
            MagicMock(
                matches=[MagicMock()],
                unique={},
            ),
            MagicMock(
                matches=[MagicMock()],
                unique={},
            ),
        ]
    )


@patch("yardstick.compare_results")
def test_validate_fail_on_empty_matches(mock_compare_results, compare_results_no_matches):
    mock_compare_results.return_value = compare_results_no_matches
    gate = validate_image(
        "some image",
        GateConfig(fail_on_empty_match_set=True),
        descriptions=["some-str", "another-str"],
        always_run_label_comparison=False,
        verbosity=0,
    )
    assert not gate.passed()
    assert "gate configured to fail on empty matches, and no matches found" in gate.reasons
    mock_compare_results.assert_called_once_with(
        descriptions=["some-str", "another-str"],
        year_max_limit=None,
        matches_filter=None,
    )


@patch("yardstick.compare_results")
def test_validate_dont_fail_on_empty_matches(mock_compare_results, compare_results_no_matches):
    mock_compare_results.return_value = compare_results_no_matches
    gate = validate_image(
        "some image",
        GateConfig(fail_on_empty_match_set=False),
        descriptions=["some-str", "another-str"],
        always_run_label_comparison=False,
        verbosity=0,
    )
    assert gate.passed()
    mock_compare_results.assert_called_once_with(
        descriptions=["some-str", "another-str"],
        year_max_limit=None,
        matches_filter=None,
    )


@patch("yardstick.compare_results")
def test_validate_pass_early_identical_match_sets(mock_compare_results, compare_results_identical_matches):
    mock_compare_results.return_value = compare_results_identical_matches
    gate = validate_image(
        "some image",
        GateConfig(fail_on_empty_match_set=False),
        descriptions=["some-str", "another-str"],
        always_run_label_comparison=False,
        verbosity=0,
    )
    assert gate.passed()
    mock_compare_results.assert_called_once_with(
        descriptions=["some-str", "another-str"],
        year_max_limit=None,
        matches_filter=None,
    )


@pytest.fixture()
def reference_config():
    return ScanConfiguration(
        image_repo="docker.io/anchore/test_images",
        image_digest="f" * 64,
        tool_name="grype",
        tool_version="123",
        tool_label="reference",
        ID="reference-config-uuid",
    )


@pytest.fixture()
def candidate_config():
    return ScanConfiguration(
        image_repo="docker.io/anchore/test_images",
        image_digest="f" * 64,
        tool_name="grype",
        tool_version="1234",
        tool_label="candidate",
        ID="candidate-config-uuid",
    )


@pytest.fixture()
def matches(packages, vulns):
    libc, nginx, openssl, zlib = packages
    vuln1, vuln2, vuln3, vuln4 = vulns
    match1 = Match(
        package=libc,
        vulnerability=vuln1,
    )
    match2 = Match(
        package=nginx,
        vulnerability=vuln2,
    )
    match3 = Match(
        package=openssl,
        vulnerability=vuln3,
    )
    match4 = Match(
        package=zlib,
        vulnerability=vuln4,
    )
    return [match1, match2, match3, match4]


@pytest.fixture()
def reference_results(reference_config, packages, matches):
    match1, match2, match3, match4 = matches
    return ScanResult(
        config=reference_config,
        matches=[match1, match2, match3],
        packages=packages,
    )


@pytest.fixture()
def candidate_results(candidate_config, packages, matches):
    match1, match2, match3, match4 = matches
    return ScanResult(
        config=candidate_config,
        matches=[match1, match2, match3, match4],
        packages=packages,
    )


@pytest.fixture()
def non_identical_results(reference_results, candidate_results):
    return comparison.ByPreservedMatch(results=[reference_results, reference_results])


@pytest.fixture()
def vulns():
    vuln1 = Vulnerability(id="CVE-2021-1234")
    vuln2 = Vulnerability(id="CVE-2021-0002")
    vuln3 = Vulnerability(id="CVE-2021-5678")
    vuln4 = Vulnerability(id="CVE-2021-8888")
    return vuln1, vuln2, vuln3, vuln4


@pytest.fixture()
def packages():
    libc = Package(name="libc", version="2.29")
    nginx = Package(name="nginx", version="1.17")
    openssl = Package(name="openssl", version="1.1.1")
    zlib = Package(name="zlib", version="1.2.11")
    return [libc, nginx, openssl, zlib]


@pytest.fixture()
def deltas():
    return [
        MagicMock(spec=Delta),
        MagicMock(spec=Delta),
    ]


@pytest.fixture()
def label_entries(matches):
    match1, match2, match3, match4 = matches
    return [
        LabelEntry(
            Label.TruePositive,
            vulnerability_id=match1.vulnerability.id,
            package=match1.package,
        ),
        LabelEntry(
            Label.FalsePositive,
            vulnerability_id=match2.vulnerability.id,
            package=match2.package,
        ),
        LabelEntry(
            Label.TruePositive,
            vulnerability_id=match3.vulnerability.id,
            package=match3.package,
        ),
        LabelEntry(
            Label.TruePositive,
            vulnerability_id=match4.vulnerability.id,
            package=match4.package,
        ),
    ]


@pytest.fixture()
def label_comparison_results(reference_results, candidate_results, label_entries):
    compare_configuration = {
        "year_max_limit": 2021,
        "year_from_cve_only": True,
    }
    return (
        [reference_results, candidate_results],
        [],  # label_entries is not used
        {
            reference_results.ID: comparison.AgainstLabels(
                result=reference_results,
                label_entries=label_entries,
                lineage=[],
                compare_configuration=compare_configuration,
            ),
            candidate_results.ID: comparison.AgainstLabels(
                result=candidate_results,
                label_entries=label_entries,
                lineage=[],
                compare_configuration=compare_configuration,
            ),
        },
        MagicMock(name="stats_by_image_tool_pair"),
    )


@patch("yardstick.compare_results")
@patch("yardstick.compare_results_against_labels")
@patch("yardstick.validate.delta.compute_deltas")
def test_validate_non_identical_match_sets(
    mock_compute_deltas,
    mock_compare_against_labels,
    mock_compare_results,
    non_identical_results,
    deltas,
    label_comparison_results,
):
    mock_compare_results.return_value = non_identical_results
    mock_compare_against_labels.return_value = label_comparison_results
    mock_compute_deltas.return_value = deltas
    gate = validate_image(
        f"docker.io/anchore/test_images@{'f' * 64}",
        GateConfig(fail_on_empty_match_set=False),
        descriptions=["some-str", "another-str"],
        always_run_label_comparison=False,
        verbosity=0,
    )
    assert gate.passed()



from dataclasses import dataclass

DEFAULT_STORE_ROOT = ".yardstick"


@dataclass
class StoreConfig:
    store_root: str = DEFAULT_STORE_ROOT


_config = StoreConfig()


def get() -> StoreConfig:
    return _config


def set_values(**kwargs):
    for k, v in kwargs.items():
        setattr(get(), k, v)



import pytest
from yardstick import store


class TestNamingImage:
    @pytest.mark.parametrize(
        ("image", "expected"),
        [
            ("ubuntu:20.04", "ubuntu:20.04"),
            ("anchore/anchore-engine:latest", "anchore+anchore-engine:latest"),
            ("something/nested/image:latest", "something+nested+image:latest"),
        ],
    )
    def test_encode_decode(self, image, expected):
        assert expected == store.naming.image.encode(image)
        assert image == store.naming.image.decode(expected)



from . import (
    config,
    image_lineage,
    label_stats,
    labels,
    naming,
    result_set,
    scan_result,
    tool,
)

__all__ = [
    "config",
    "image_lineage",
    "label_stats",
    "labels",
    "naming",
    "result_set",
    "scan_result",
    "tool",
]



import logging
import sys

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


def load_plugins():
    tool_plugins = entry_points(group="yardstick.plugins.tools")
    logging.debug(f"discovered plugin entrypoints: {tool_plugins}")

    for tool in tool_plugins:
        try:
            logging.info(f"Loading tool plugin {tool.name}")
            tool.load()
        except:  # noqa: E722
            logging.exception(f"Failed loading tool plugin {tool.name}")



import logging
from typing import Callable, Optional

from . import (
    arrange,
    artifact,
    capture,
    cli,
    comparison,
    label,
    store,
    tool,
    validate,
    utils,
)

__all__ = [
    "arrange",
    "artifact",
    "capture",
    "cli",
    "comparison",
    "label",
    "store",
    "tool",
    "validate",
    "utils",
]


def compare_results(
    descriptions: list[str],
    year_max_limit: Optional[int] = None,
    year_from_cve_only: bool = False,
    matches_filter: Optional[Callable] = None,
    store_root: Optional[str] = None,
) -> comparison.ByPreservedMatch:
    results = store.scan_result.load_by_descriptions(
        descriptions=descriptions,
        year_max_limit=year_max_limit,
        year_from_cve_only=year_from_cve_only,
        skip_sbom_results=True,
        store_root=store_root,
    )

    if matches_filter:
        for result in results:
            result.matches = matches_filter(result.matches)

    digests = {r.config.image_digest for r in results}
    if len(digests) != 1:
        raise RuntimeError(f"image digests being compared do not match: {digests}")

    return comparison.ByPreservedMatch(results=results)


def compare_results_against_labels(  # noqa: PLR0913
    descriptions: list[str],
    result_set: Optional[str] = None,
    fuzzy: bool = False,
    year_max_limit: Optional[int] = None,
    year_from_cve_only: bool = False,
    label_entries: Optional[list[artifact.LabelEntry]] = None,
    matches_filter: Optional[Callable] = None,
    store_root: Optional[str] = None,
) -> tuple[
    list[artifact.ScanResult],
    list[artifact.LabelEntry],
    dict[str, comparison.AgainstLabels],
    comparison.ImageToolLabelStats,
]:
    descriptions = list(descriptions)

    # this is a description of what was done to the results before comparison
    # we want to keep this on the comparison to evaluate if the comparison result
    # can be compared to other comparison results (are they apples to apples).
    compare_configuration = {
        "year_max_limit": year_max_limit,
        "year_from_cve_only": year_from_cve_only,
    }

    if result_set:
        descriptions.extend(store.result_set.load(result_set).descriptions)

    if not descriptions:
        raise RuntimeError("no descriptions provided")

    logging.debug(f"running label comparison with {descriptions}")

    results = store.scan_result.load_by_descriptions(
        descriptions,
        skip_sbom_results=True,
        year_max_limit=year_max_limit,
        year_from_cve_only=year_from_cve_only,
        store_root=store_root,
    )

    if matches_filter:
        for result in results:
            result.matches = matches_filter(result.matches)

    if label_entries is None:
        label_entries = store.labels.load_all(
            year_max_limit=year_max_limit,
            year_from_cve_only=year_from_cve_only,
            store_root=store_root,
        )

    (
        comparisons_by_result_id,
        stats_by_image_tool_pair,
    ) = comparison.of_results_against_label(
        *results,
        fuzzy_package_match=fuzzy,
        label_entries=label_entries,
        compare_configuration=compare_configuration,
    )

    return results, label_entries, comparisons_by_result_id, stats_by_image_tool_pair


def compare_results_against_labels_by_ecosystem(
    result_set: str,
    fuzzy: bool = False,
    year_max_limit: Optional[int] = None,
    year_from_cve_only: bool = False,
    label_entries: Optional[list[artifact.LabelEntry]] = None,
) -> tuple[
    dict[str, list[artifact.ScanResult]],
    list[artifact.LabelEntry],
    comparison.ToolLabelStatsByEcosystem,
]:
    results = store.result_set.load_scan_results(
        result_set,
        year_max_limit=year_max_limit,
        skip_sbom_results=True,
    )
    results_by_image = arrange.scan_results_by_image(results)

    if label_entries is None:
        label_entries = store.labels.load_all(
            year_max_limit=year_max_limit,
            year_from_cve_only=year_from_cve_only,
        )

    stats = comparison.of_results_against_label_by_ecosystem(
        results_by_image,
        fuzzy_package_match=fuzzy,
        label_entries=label_entries,
    )
    return results_by_image, label_entries, stats



from yardstick import artifact, comparison

# TODO: include test with lineage


def test_comparison_against_labels():
    config = artifact.ScanConfiguration(
        image_repo="myimage",
        image_digest="123456",
        tool_name="grype",
        tool_version="main",
    )

    package_bash_5 = artifact.Package(name="bash", version="5.0-6ubuntu1.1")
    package_coreutils_8 = artifact.Package(name="coreutils", version="8.30-3ubuntu2")
    package_libc_2 = artifact.Package(name="libc-bin", version="2.31-0ubuntu9.2")
    package_libsystemd_245 = artifact.Package(
        name="libsystemd0",
        version="245.4-4ubuntu3.2",
    )
    package_libsystemd_25 = artifact.Package(name="libsystemd0", version="25")

    expected_tp_matches = [
        artifact.Match(
            vulnerability=artifact.Vulnerability(id="CVE-2019-18276"),
            package=package_bash_5,
            config=config,
        ),
        artifact.Match(
            vulnerability=artifact.Vulnerability(id="CVE-2016-2781"),
            package=package_coreutils_8,
            config=config,
        ),
    ]

    expected_fp_matches = [
        artifact.Match(
            vulnerability=artifact.Vulnerability(id="CVE-2016-10228"),
            package=package_libc_2,
            config=config,
        ),
        artifact.Match(
            vulnerability=artifact.Vulnerability(id="CVE-2018-20839"),
            package=package_libsystemd_245,
            config=config,
        ),
        artifact.Match(
            vulnerability=artifact.Vulnerability(id="CVE-2018-29999"),
            package=package_libsystemd_25,
            config=config,
        ),
    ]

    matches = [
        *expected_tp_matches,
        *expected_fp_matches,
    ]

    result = artifact.ScanResult(
        config=config,
        matches=matches,
    )

    common_label_options = {
        "image": artifact.ImageSpecifier(exact=config.image),
        "source": "manual",
    }

    false_negative_label_entries = [
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="CVE-2016-NM111",
            package=package_libc_2,
            **common_label_options,
        ),
        # do not have matches and already are covered with effective CVE... but the package mismatches
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="ELSA-2020-123567",
            effective_cve="CVE-2019-18276",
            package=package_libsystemd_25,
            **common_label_options,
        ),
    ]

    label_entries = [
        # have matches
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="CVE-2019-18276",
            package=package_bash_5,
            **common_label_options,
        ),
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="CVE-2016-2781",
            package=package_coreutils_8,
            **common_label_options,
        ),
        artifact.LabelEntry(
            label=artifact.Label.FalsePositive,
            vulnerability_id="CVE-2016-10228",
            package=package_libc_2,
            **common_label_options,
        ),
        artifact.LabelEntry(
            label=artifact.Label.FalsePositive,
            vulnerability_id="CVE-2018-20839",
            package=package_libsystemd_245,
            **common_label_options,
        ),
        artifact.LabelEntry(
            label=artifact.Label.FalsePositive,
            vulnerability_id="CVE-2018-29999",
            package=package_libsystemd_25,
            **common_label_options,
        ),
        # do not have matches
        *false_negative_label_entries,
        artifact.LabelEntry(
            label=artifact.Label.FalsePositive,
            vulnerability_id="CVE-2016-NM222",
            package=package_libc_2,
            **common_label_options,
        ),
        artifact.LabelEntry(
            label=artifact.Label.FalsePositive,
            vulnerability_id="CVE-2016-NM333",
            package=package_libsystemd_25,
            **common_label_options,
        ),
        # do not have matches and already are covered with effective CVE
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="ELSA-2020-123567",
            effective_cve="CVE-2019-18276",
            package=package_bash_5,
            **common_label_options,
        ),
    ]

    actual = comparison.AgainstLabels(
        result=result,
        label_entries=label_entries,
        lineage=[],
    )

    assert actual.summary.true_positives == len(expected_tp_matches)
    assert actual.summary.false_positives == len(expected_fp_matches)
    assert actual.summary.false_negatives == len(false_negative_label_entries)
    assert actual.summary.f1_score == 0.4444444444444444

    assert set(actual.true_positive_matches) == set(expected_tp_matches)
    assert set(actual.false_positive_matches) == set(expected_fp_matches)
    assert set(actual.false_negative_label_entries) == set(false_negative_label_entries)


def test_comparison_against_labels_indeterminate():
    config = artifact.ScanConfiguration(
        image_repo="myimage",
        image_digest="123456",
        tool_name="grype",
        tool_version="main",
    )

    package_bash_5 = artifact.Package(name="bash", version="5.0-6ubuntu1.1")

    m1 = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2019-18276"),
        package=package_bash_5,
        config=config,
    )
    m2 = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2020-12000"),
        package=package_bash_5,
        config=config,
    )
    m3 = artifact.Match(
        vulnerability=artifact.Vulnerability(id="CVE-2020-2222"),
        package=package_bash_5,
        config=config,
    )

    matches = [m1, m2, m3]

    result = artifact.ScanResult(
        config=config,
        matches=matches,
    )

    common_label_options = {
        "image": artifact.ImageSpecifier(exact=config.image),
        "source": "manual",
    }

    m2_indeterminate = [
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="CVE-2020-12000",
            package=package_bash_5,
            **common_label_options,
        ),
        artifact.LabelEntry(
            label=artifact.Label.FalsePositive,
            vulnerability_id="CVE-2020-12000",
            package=package_bash_5,
            **common_label_options,
        ),
    ]

    m3_indeterminate = [
        artifact.LabelEntry(
            label=artifact.Label.Unclear,
            vulnerability_id="CVE-2020-2222",
            package=package_bash_5,
            **common_label_options,
        ),
    ]

    label_entries = [
        # not indeterminate
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="CVE-2019-18276",
            package=package_bash_5,
            **common_label_options,
        ),
        artifact.LabelEntry(
            label=artifact.Label.TruePositive,
            vulnerability_id="CVE-2019-18276",
            package=package_bash_5,
            **common_label_options,
        ),
        # indeterminate
        *m2_indeterminate,
        *m3_indeterminate,
    ]

    actual = comparison.AgainstLabels(
        result=result,
        label_entries=label_entries,
        lineage=[],
    )

    assert actual.summary.indeterminate == 1
    assert set(actual.matches_with_indeterminate_labels) == {m2}
    assert actual.summary.f1_score == 1
    assert actual.summary.f1_score_lower_confidence == 0.6666666666666666
    assert actual.summary.f1_score_upper_confidence == 1



from asyncio import Future
from typing import List, Optional, Union

from prompt_toolkit.formatted_text import (
    AnyFormattedText,
    merge_formatted_text,
    to_formatted_text,
)
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import (
    ConditionalContainer,
    Container,
    Dimension,
    FormattedTextControl,
)
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.margins import ScrollbarMargin

from yardstick import artifact
from yardstick.cli.explore.image_labels.edit_note_dialog import EditNoteDialog
from yardstick.cli.explore.image_labels.label_json_editor_dialog import (
    LabelJsonEditorDialog,
)
from yardstick.cli.explore.image_labels.label_manager import LabelManager


class LabelSelectionPane:
    def __init__(
        self,
        label_manager: LabelManager,
        dialog_executor,
        filter=None,  # noqa: A002
    ) -> None:
        self.label_manager = label_manager
        self.dialog_executor = dialog_executor
        self.entries: List[artifact.LabelEntry] = []
        self.selected_entry = 0
        self.match: Optional[artifact.Match] = None
        # self.width = 80
        self.container: Union[HSplit, ConditionalContainer] = HSplit(
            [
                Window(
                    content=FormattedTextControl(
                        text=to_formatted_text(
                            "Match Label Details",
                            style="bold reverse",
                        ),
                        focusable=False,
                    ),
                    style="class:pane-title",
                    height=Dimension.exact(1),
                    cursorline=True,
                    always_hide_cursor=True,
                    wrap_lines=False,
                ),
                Window(
                    content=FormattedTextControl(
                        text=self.get_formatted_result_list,
                        focusable=True,
                        key_bindings=self._get_key_bindings(),
                    ),
                    style="class:select-label-box",
                    # width=Dimension.exact(self.width),
                    height=Dimension(preferred=15),
                    cursorline=True,
                    always_hide_cursor=False,
                    wrap_lines=True,
                    right_margins=[
                        ScrollbarMargin(display_arrows=False),
                    ],
                ),
            ],
        )

        if filter is not None:
            self.container = ConditionalContainer(content=self.container, filter=filter)

    def set_match(self, match: artifact.Match):
        self.selected_entry = 0
        self.match = match
        self._update()

    def _update(self):
        self.entries = self.label_manager.get_label_entries_by_match(match=self.match)

    def get_formatted_result_list(self) -> AnyFormattedText:
        result: List[AnyFormattedText] = []

        for i, entry in enumerate(self.entries):
            if i == self.selected_entry:
                result.append([("[SetCursorPosition]", "")])

            note = entry.note
            formatted_note = to_formatted_text("[no note provided]", style="italic") if not note else to_formatted_text(note)

            label_style = ""
            if entry.label == artifact.Label.TruePositive:
                label_style = "#428bff"
            elif entry.label == artifact.Label.FalsePositive:
                label_style = "#ff0066"
            elif entry.label == artifact.Label.Unclear:
                label_style = "#888888"

            result += [
                to_formatted_text(f"{i + 1}. "),
                to_formatted_text(entry.label.display, label_style),
                to_formatted_text(f" [{entry.ID}]"),
                to_formatted_text(f" {entry.vulnerability_id}"),
            ]

            result += [
                to_formatted_text(f" from {entry.user}"),
                ": ",
                formatted_note,
                "\n",
                # "\n" + "━" * (self.width - 1) + "\n",
            ]

        if not result:
            result.append(to_formatted_text("[no labels for match]", style="italic"))
        return merge_formatted_text(result)

    def get_selected_entry(self):
        if self.selected_entry > len(self.entries) - 1:
            return None
        return self.entries[self.selected_entry]

    def _get_key_bindings(self) -> KeyBindings:  # noqa: C901
        kb = KeyBindings()

        @kb.add("up")
        def _go_up(event) -> None:
            if self.entries:
                self.selected_entry = (self.selected_entry - 1) % len(self.entries)

        @kb.add("down")
        def _go_down(event) -> None:
            if self.entries:
                self.selected_entry = (self.selected_entry + 1) % len(self.entries)

        @kb.add("backspace")
        @kb.add("delete")
        def _delete_entry(event) -> None:
            entry = self.get_selected_entry()
            if not entry:
                return

            self.label_manager.remove_label_entry(entry.ID)

            self._update()

        @kb.add("e")
        def _edit_note(event) -> None:
            entry = self.get_selected_entry()
            if not entry:
                return

            def done(fut: Future):
                if fut.done() and fut.result() is not None:
                    note = fut.result()
                    self.label_manager.edit_label_entry_note(entry.ID, note)
                    self._update()

            self.dialog_executor(
                dialog=EditNoteDialog(
                    title="Edit note",
                    label_text="Note:",
                    text_area=True,
                    ok_button_text="Update",
                    default_value=entry.note,
                ),
                done_callback=done,
            )

        @kb.add("j")
        def _edit_json(event) -> None:
            entry = self.get_selected_entry()
            if not entry:
                return

            def done(fut: Future):
                if fut.done() and fut.result() is not None:
                    self.label_manager.edit_label_entry_json(entry.ID, fut.result())
                    self._update()

            self.dialog_executor(
                dialog=LabelJsonEditorDialog(entry),
                done_callback=done,
            )

        return kb

    def __pt_container__(self) -> Container:
        return self.container



from __future__ import annotations

import contextlib
import hashlib
import json
import logging
import os
import shutil
from typing import Any

from yardstick import comparison
from yardstick.store import config as store_config

LABEL_STATS_DIR = "label-stats"


def _store_root(store_root: str | None = None) -> str:
    if not store_root:
        store_root = store_config.get().store_root
    return os.path.join(store_root, LABEL_STATS_DIR)


def store_path(
    ids: list[str],
    configuration: list[dict[str, Any]] | None,
    store_root: str | None = None,
) -> str:
    config_str = _configuration_string(configuration)

    filename = "_".join(sorted(ids)) + "_" + config_str + ".json"

    return os.path.join(_store_root(store_root=store_root), filename)


def clear(store_root: str | None = None):
    with contextlib.suppress(FileNotFoundError):
        shutil.rmtree(_store_root(store_root=store_root))


def save(result: comparison.ImageToolLabelStats, store_root: str | None = None):
    if not isinstance(result, comparison.ImageToolLabelStats):
        raise RuntimeError(
            f"only ImageToolLabelStats is supported, given {type(result)}",
        )

    ids = [c.ID for c in result.configs]
    path = store_path(ids, result.compare_configs, store_root=store_root)
    logging.debug(f"storing label comparison state for {ids!r}")

    os.makedirs(os.path.dirname(path), exist_ok=True)

    with open(path, "w", encoding="utf-8") as data_file:
        data_file.write(json.dumps(result.to_dict(), indent=2))  # type: ignore[attr-defined]


def _configuration_string(configurations: list[dict[str, str]] | None) -> str:
    if configurations is None:
        return "no-configuration"

    config_strs = []
    for configuration in configurations:
        config_str = ",".join([f"{k}={v}" for k, v in sorted(configuration.items())])
        config_strs.append(config_str)

    return hashlib.sha256(",".join(config_strs).encode("utf-8")).hexdigest()


def load(
    ids: list[str],
    configurations: list[dict[str, Any]] | None = None,
    store_root: str | None = None,
) -> comparison.ImageToolLabelStats:
    data_path = store_path(ids, configurations, store_root=store_root)

    logging.debug(
        f"loading label comparison state for {ids!r} with detailed configurations location={data_path!r}",
    )

    with open(data_path, encoding="utf-8") as data_file:
        data_json = data_file.read()

    return comparison.ImageToolLabelStats.from_json(  # type: ignore[attr-defined]
        data_json,
    )



import collections
import datetime
import functools
import getpass
import hashlib
import json
import os
import re
import subprocess
import uuid
from dataclasses import InitVar, asdict, dataclass, field
from enum import Enum
from typing import Any

from dataclasses_json import config, dataclass_json

from yardstick.utils import grype_db, is_cve_vuln_id, parse_year_from_id


def get_image_digest(image: str) -> str:
    result = subprocess.run(
        ["docker", "manifest", "inspect", image],
        stdout=subprocess.PIPE,
        check=False,
    )
    if result.returncode != 0:
        raise RuntimeError(f"failed to get image digest for {image}")
    obj = json.loads(result.stdout)
    digest = obj.get("config", {}).get("digest", "")
    if digest:
        return digest
    for manifest in obj.get("manifests", []):
        if manifest.get("platform", {}) != {"architecture": "amd64", "os": "linux"}:
            continue
        digest = manifest.get("digest", "")
        if digest:
            return digest
    raise RuntimeError(f"no image digest found for {image}")


@dataclass(frozen=True, eq=True)
class Tool:
    tool: str
    name: str | None = field(init=False)
    version: str = field(init=False)
    label: str = None

    def __post_init__(self):
        name, version = self.tool.split("@", 1)
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "version", version)

    def __str__(self):
        return self.id_version

    @property
    def id_version(self):
        return f"{self.id}@{self.version}"

    @property
    def id(self):  # noqa: A003
        if self.label:
            return f"{self.name}[{self.label}]"
        return self.name

    def __lt__(self, other):
        if not isinstance(other, Tool):
            return NotImplemented
        return self.id < other.id


@dataclass()
class Image:
    image: str
    repository: str = field(init=False)
    tag: str = field(init=False)
    digest: str = field(init=False)

    def __post_init__(self):
        digest = ""
        repo = self.image
        if "@" in repo:
            repo, digest = repo.split("@", 1)

        tag = ""
        if ":" in repo:
            repo, tag = repo.split(":", 1)

        self.repository = repo
        self.tag = tag
        self.digest = digest

    @property
    def repository_encoded(self) -> str:
        return self.repository.replace("/", "+")

    @property
    def encoded(self) -> str:
        return f"{self.repository_encoded}@{self.digest}"

    def is_like(self, other: str) -> bool:
        other_obj = Image(other)
        if self.repository != other_obj.repository:
            return False
        if self.tag and other_obj.tag and self.tag != other_obj.tag:
            return False
        if self.digest and other_obj.digest and self.digest != other_obj.digest:
            return False
        return True


# note: this dataclass cannot be modified as it is used in Match.id and responsible for accessing previously stored results.
# note: we cannot freeze this class since the tool installation may add additional metadata
@dataclass_json
@dataclass(frozen=False, eq=True, order=True)
class ScanConfiguration:
    image_repo: str
    image_digest: str
    tool_name: str
    tool_version: str
    tool_label: str | None = None
    image_tag: str = ""
    timestamp: datetime.datetime | None = field(
        default=None,
        metadata=config(
            encoder=lambda dt: dt.isoformat(),
            decoder=datetime.datetime.fromisoformat,
        ),
    )
    tool_input: str | None = None
    detail: dict[str, dict | str] = field(default_factory=dict)
    ID: str = field(default_factory=lambda: str(uuid.uuid4()))

    def __post_init__(self):
        self.image_repo = self.image_repo.replace("+", "/")

    @property
    def raw_tool_name(self):
        # given anything like "grype[label]" return "grype"
        if "[" in self.tool_name:
            return self.tool_name.split("[", 1)[0]

        # given anything like "grype@v0.30.0+linux" return "grype"
        if "@" in self.tool_name:
            return self.tool_name.split("@", 1)[0]

        return self.tool_name

    @property
    def path(self):
        return f"{self.image_repo}@{self.image_digest}/{self.tool}/{self.timestamp_rfc3339}"

    @property
    def tool(self):
        return f"{self.tool_name}@{self.tool_version}"

    @property
    def encoded_path(self):
        return f"{self.image_encoded}/{self.tool}/{self.timestamp_rfc3339}"

    @property
    def image_repo_encoded(self) -> str:
        return self.image_repo.replace("/", "+")

    @property
    def timestamp_rfc3339(self) -> str:
        return self.timestamp.isoformat() if self.timestamp else ""

    @property
    def image(self):
        return f"{self.image_repo}@{self.image_digest}"

    @property
    def full_image(self):
        if self.image_tag == "":
            return self.image

        return f"{self.image_repo}:{self.image_tag}@{self.image_digest}"

    @property
    def image_encoded(self):
        return f"{self.image_repo_encoded}@{self.image_digest}"

    def __str__(self):
        s = f"""\
image:\t\t{self.image_repo}{":" + self.image_tag if self.image_tag else ""}@{self.image_digest}
tool:\t{self.tool}"""
        if self.timestamp:
            s += f"\ntimestamp:\t{self.timestamp}"
        return s

    @staticmethod
    def new(
        image: str | None = None,
        tool: str | None = None,
        path: str | None = None,
        timestamp: datetime.datetime | None = None,
        label: str | None = None,
    ):
        if tool:
            tool_obj = Tool(tool, label=label)

        if image:
            img = Image(image)
            if img.digest == "":
                img.digest = get_image_digest(image)

        if path:
            image_and_digest, tool_and_version, timestamp_str = path.rsplit("/", 2)
            tool_obj = Tool(tool_and_version)
            timestamp = datetime.datetime.fromisoformat(timestamp_str)
            img = Image(image_and_digest)

        return ScanConfiguration(
            image_repo=img.repository,
            image_digest=img.digest,
            image_tag=img.tag,
            tool_name=tool_obj.id,
            tool_version=tool_obj.version,
            timestamp=timestamp,
            tool_label=label,
        )


@dataclass(frozen=True, eq=True, order=True)
class ScanMetadata:
    timestamp: datetime.datetime | None = field(
        default=None,
        metadata=config(
            encoder=lambda dt: dt.isoformat(),
            decoder=datetime.datetime.fromisoformat,
        ),
    )
    elapsed: float | None = field(default=None)
    image_digest: str | None = field(default=None)


@dataclass(frozen=True, eq=True, order=True)
class Package:
    name: str
    version: str

    def __repr__(self):
        return f"{self.name}@{self.version}"


@dataclass(frozen=True, eq=True, order=True)
class Vulnerability:
    id: str  # noqa: A003
    cve_id: str | None = field(default=None, hash=False)

    def __post_init__(self):
        if is_cve_vuln_id(self.id) and not self.cve_id:
            object.__setattr__(self, "cve_id", self.id)

    def _get_cve(self):
        cve = grype_db.normalize_to_cve(self.id)
        if is_cve_vuln_id(cve):
            return cve
        return None

    def _effective_cve_year(self) -> int | None:
        cve = self.cve_id
        if not cve:
            # this is rather expensive, so try this last
            cve = self._get_cve()
        if not cve:
            return None
        return parse_year_from_id(cve)

    def effective_year(self, by_cve=False) -> int | None:
        if by_cve:
            return self._effective_cve_year()
        candidate: int | str | None = self.id
        if self.id:
            candidate = parse_year_from_id(self.id)
        if not isinstance(candidate, int):
            candidate = self._effective_cve_year()
        if isinstance(candidate, int):
            return candidate
        return None


class DTEncoder(json.JSONEncoder):
    def default(self, o):
        # if passed in object is datetime object convert it to a string
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        # otherwise use the default behavior
        return json.JSONEncoder.default(self, o)


# note: cannot use order=True since lt/gt/etc require not including the fullentry dict
@dataclass_json
@dataclass(frozen=True)
class Match:
    vulnerability: Vulnerability
    package: Package
    fullentry: dict[str, Any] | None = field(default=None, hash=False)
    config: ScanConfiguration | None = field(default=None, hash=False)
    ID: str = field(init=False, hash=False)

    def __post_init__(self):
        identifier = {
            "output": self.fullentry,
            "configuration": None,
        }
        if self.config:
            identifier["configuration"] = sorted(
                asdict(self.config).items(),
            )

        # note on S324 usage: we are not using these IDs for crytographic purposes, only
        # for having simple content-sensitive IDs for use for match comparison.
        match_id = hashlib.md5(  # noqa: S324
            json.dumps(identifier, sort_keys=True, cls=DTEncoder).encode(),
        ).hexdigest()

        object.__setattr__(self, "ID", match_id)

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def __repr__(self) -> str:
        # note: fullentry is excluded
        return f"Match(vulnerability={self.vulnerability.id!r} {self.package!r} id={self.ID!r})"

    def __lt__(self, other: Any) -> bool:
        # don't compare the full entry
        if self.package.name.lower() != other.package.name.lower():
            return self.package.name.lower() < other.package.name.lower()

        if self.vulnerability.id.lower() != other.vulnerability.id.lower():
            return self.vulnerability.id.lower() < other.vulnerability.id.lower()

        if self.package.version != other.package.version:
            return self.package.version < other.package.version

        if self.config and other.config and self.package.version != other.package.version:
            return self.package.version < other.package.version
        return False


@dataclass_json
@dataclass(frozen=False, eq=True, order=True)
class ScanResult:
    config: ScanConfiguration
    matches: list[Match] | None = field(default=None)
    packages: list[Package] | None = field(default=None)
    metadata: ScanMetadata | None = field(default=None)

    @property
    def ID(self):
        return self.config.ID

    def __post_init__(self):
        if self.matches is None and self.packages is None:
            raise RuntimeError("must have at either matches or packages (or both)")


class Label(Enum):
    TruePositive = "TP"
    FalsePositive = "FP"
    Unclear = "??"

    def __init__(self, display: str):
        self.display = display

    @staticmethod
    def encode(label):
        return label.display

    @staticmethod
    def decode(val):
        return Label.from_str(val)

    @staticmethod
    def from_str(text: str):
        text = text.lower()
        if text in ("tp", "true", "truepositive", "true-positive"):
            return Label.TruePositive
        if text in ("fp", "false", "falsepositive", "false-positive"):
            return Label.FalsePositive
        if text in (
            "unclear",
            "uncertain",
            "indeterminate",
            "dunno",
            "?",
            "??",
            "?!",
            r"¯\_(ツ)_/¯",
        ):
            return Label.Unclear
        return None


@dataclass_json
@dataclass(frozen=True, eq=True, order=True)
class ImageSpecifier:
    exact: str | None = None
    regex: str | None = None
    prefix: str | None = None
    suffix: str | None = None

    def matches_image(self, image: str):
        if not self.exact and not self.regex and not self.prefix and not self.suffix:
            # if the specification is empty then any image matches automatically
            return True
        if self.exact and image == self.exact:
            return True

        if self.prefix and image.startswith(self.prefix):
            return True

        if self.suffix and image.endswith(self.suffix):
            return True

        if self.regex and re.match(self.regex, image):
            return True
        return False

    def __repr__(self):
        result = ""
        if self.exact:
            result += self.exact

        if self.regex:
            result += f" regex={self.regex!r}"

        if self.prefix:
            result += f" prefix={self.prefix!r}"

        if self.suffix:
            result += f" suffix={self.suffix!r}"

        return result


@dataclass_json
@dataclass(eq=True, order=True)
class LabelEntry:
    # label: Label
    label: Label = field(
        metadata=config(
            encoder=Label.encode,
        ),
    )  # TP/FP/Unclear indication (required)
    vulnerability_id: str  # the CVE ID (required)
    image: ImageSpecifier | None = None  # image specifier
    note: str | None = None  # a general comment field (optional)
    source: str | None = None  # e.g. manually added, import, etc. (optional)
    effective_cve: str | None = None  # the CVE the vulnerability ID matches to (optional)
    user: str | None = None  # the user that added this label (optional)
    package: Package | None = None  # package name/version
    fullentry_fields: list[str] = field(
        default_factory=list,
    )  # values that must be found in the full_entry field on the match object
    timestamp: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now(tz=datetime.timezone.utc),
        metadata=config(
            encoder=lambda dt: dt.isoformat(),
            decoder=datetime.datetime.fromisoformat,
        ),
    )
    tool: str | None = None  # used to indicate the tool that first saw this label
    lookup_effective_cve: InitVar[bool] = False

    ID: str = field(default_factory=lambda: str(uuid.uuid4()))

    def __post_init__(self, lookup_effective_cve: bool):
        if not self.user:
            self.user = getpass.getuser()
        if lookup_effective_cve and not self.effective_cve:
            self.effective_cve = self._get_cve()

    def __str__(self):
        return f"""\
label: {self.label.name}
vulnerability: {self.vulnerability_id}
package: {self.package}
user: {self.user}
image: {self.image}
number of fullentry_fields: {len(self.fullentry_fields)}
timestamp: {self.timestamp.isoformat()}
id: {self.ID}
"""

    def __hash__(self) -> int:
        # dont include user, ID, or datetime
        return hash(
            (
                self.vulnerability_id,
                self.source,
                self.image,
                self.package,
                self.note,
                # note: we cannot depend on the dataclasses generated hash since fullentry_fields must be hashable
                tuple(sorted(self.fullentry_fields)),
            ),
        )

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def matches_image(self, image: str | None):
        if not self.image or not image:
            # if no imag specifier is provided, then any image matches automatically
            return True
        return self.image.matches_image(image)

    def summarize(self):
        return f"[{self.ID}] {self.label.name} claim {self.vulnerability_id} on package {self.package or '(any package)'} for {self.image}"

    def _get_cve(self):
        cve = grype_db.normalize_to_cve(self.vulnerability_id)
        if is_cve_vuln_id(cve):
            return cve
        return None

    def _effective_cve_year(self) -> int | None:
        cve = self.effective_cve
        if not cve:
            # this is rather expensive, so try this last
            cve = self._get_cve()
        if not cve:
            return None
        return parse_year_from_id(cve)

    def effective_year(self, by_cve=False) -> int | None:
        if by_cve:
            return self._effective_cve_year()
        candidate: int | str | None = self.vulnerability_id
        if self.vulnerability_id:
            candidate = parse_year_from_id(self.vulnerability_id)
        if not isinstance(candidate, int):
            candidate = self._effective_cve_year()
        if isinstance(candidate, int):
            return candidate
        return None


class LabelEntryCollection:
    def __init__(self, entries: list[LabelEntry]):
        self.entries = entries

    @functools.cache  # noqa: B019
    def for_image(self, image: str) -> list[LabelEntry]:
        return [e for e in self.entries if e.matches_image(image)]


@dataclass()
class ScanRequest:
    image: str
    tool: str
    label: str | None = None
    profile: str | None = None
    provides: str | None = None
    takes: str | None = None
    refresh: bool = True

    @staticmethod
    def needs_rendering(tool: str) -> bool:
        return "@env:" in tool or "@git:current-commit" in tool

    @staticmethod
    def render_tool(tool: str) -> str:
        if "@git:current-commit" in tool:
            val = (
                subprocess.check_output(
                    ["git", "rev-parse", "HEAD"],
                )
                .decode("ascii")
                .strip()
            )
            # preserve the name and any other suffix
            tool = tool.replace("@git:current-commit", f"@{val}")

        if "@env:" in tool:
            name, val = tool.split("@env:")
            if "+" in val:
                val, metadata = val.split("+", 1)
            else:
                metadata = None
            # preserve the name and any other suffix
            tool = f"{name}@{os.environ[val]}"
            if metadata:
                tool += f"+{metadata}"

        return tool

    def __post_init__(self):
        if self.needs_rendering(self.tool):
            object.__setattr__(self, "tool", self.render_tool(self.tool))


@dataclass_json
@dataclass()
class ResultState:
    request: ScanRequest
    config: ScanConfiguration | None = None


@dataclass_json
@dataclass()
class ResultSet:
    name: str
    state: list[ResultState] = field(default_factory=list)

    def get(self, tool: str, image: str) -> ResultState | None:
        img = Image(image)
        for state in self.state:
            if state.request.tool == tool and img.is_like(state.request.image):
                return state
        return None

    def provider(self, image: str, provides: str) -> ResultState | None:
        img = Image(image)
        for state in self.state:
            if img.is_like(state.request.image) and state.request.provides == provides:
                return state
        return None

    def add(self, request: ScanRequest, scan_config: ScanConfiguration):
        self.state.append(ResultState(request=request, config=scan_config))

    @property
    def descriptions(self) -> list[str]:
        descriptions = []
        for result_state in self.state:
            if result_state.config:
                descriptions.append(result_state.config.path)
        return descriptions

    @property
    def ids(self) -> list[str]:
        ids = []
        for result_state in self.state:
            if result_state.config:
                ids.append(result_state.config.ID)
        return ids

    @property
    def result_state_by_image(self) -> dict[str, list[ResultState]]:
        descriptions = collections.defaultdict(list)
        for result_state in self.state:
            if result_state.config:
                descriptions[result_state.config.image].append(result_state)
        return descriptions



from yardstick.validate import Gate, GateConfig, GateInputDescription, Delta
from yardstick import artifact, comparison


import pytest
from unittest.mock import MagicMock


@pytest.fixture
def mock_label_comparison():
    """Fixture to create a mock LabelComparisonSummary with defaults."""
    summary = MagicMock()
    summary.f1_score = 0.9
    summary.false_negatives = 5
    summary.indeterminate_percent = 2.0
    return summary


@pytest.mark.parametrize(
    "config, reference_summary, candidate_summary, expected_reasons",
    [
        # Case 1: Candidate has a lower F1 score beyond the allowed threshold -> gate fails
        (
            GateConfig(
                max_f1_regression=0.1,
                max_new_false_negatives=5,
                max_unlabeled_percent=10,
            ),
            MagicMock(f1_score=0.9, false_negatives=5, indeterminate_percent=2.0),
            MagicMock(f1_score=0.7, false_negatives=5, indeterminate_percent=2.0),
            ["current F1 score is lower than the latest release F1 score"],
        ),
        # Case 2: Candidate has too many false negatives -> gate fails
        (
            GateConfig(
                max_f1_regression=0.1,
                max_new_false_negatives=1,
                max_unlabeled_percent=10,
            ),
            MagicMock(f1_score=0.9, false_negatives=5, indeterminate_percent=2.0),
            MagicMock(f1_score=0.85, false_negatives=7, indeterminate_percent=2.0),
            ["current false negatives is greater than the latest release false negatives"],
        ),
        # Case 3: Candidate has too high indeterminate percent -> gate fails
        (
            GateConfig(
                max_f1_regression=0.1,
                max_new_false_negatives=5,
                max_unlabeled_percent=5,
            ),
            MagicMock(f1_score=0.9, false_negatives=5, indeterminate_percent=2.0),
            MagicMock(f1_score=0.85, false_negatives=5, indeterminate_percent=6.0),
            ["current indeterminate matches % is greater than"],
        ),
        # Case 4: Candidate passes all thresholds -> gate passes (no reasons)
        (
            GateConfig(
                max_f1_regression=0.1,
                max_new_false_negatives=5,
                max_unlabeled_percent=10,
            ),
            MagicMock(f1_score=0.9, false_negatives=5, indeterminate_percent=2.0),
            MagicMock(f1_score=0.85, false_negatives=5, indeterminate_percent=3.0),
            [],
        ),
    ],
)
def test_gate(config, reference_summary, candidate_summary, expected_reasons):
    """Parameterized test for the Gate class that checks different pass/fail conditions."""

    # Create the Gate instance with the given parameters
    gate = Gate(
        reference_comparison=reference_summary,
        candidate_comparison=candidate_summary,
        config=config,
        input_description=MagicMock(image="test_image"),
    )

    # Check that the reasons list matches the expected outcome
    assert len(gate.reasons) == len(expected_reasons)
    for reason, expected_reason in zip(gate.reasons, expected_reasons):
        assert expected_reason in reason


def test_gate_failing():
    input_description = GateInputDescription(image="some-image", configs=[])
    gate = Gate.failing(["sample failure reason"], input_description)
    assert not gate.passed()
    assert gate.reasons == ["sample failure reason"]


def test_gate_passing():
    input_description = GateInputDescription(image="some-image", configs=[])
    gate = Gate.passing(input_description)
    assert gate.passed()



import pytest

from yardstick.cli import config


def test_config(tmp_path):
    profile_file = tmp_path / ".yardstick.profiles.yaml"
    subject = f"""
store_root: .
profile_path: {profile_file}
default-max-year: 2021

x-ref:
  full-label-set-images: &full-label-set-images
    - docker.io/cloudbees/cloudbees-core-agent:2.289.2.2@sha256:d48f0546b4cf5ef4626136242ce302f94a42751156b7be42f4b1b75a66608880
    - docker.io/cloudbees/cloudbees-core-mm:2.277.3.1@sha256:4c564f473d38f23da1caa48c4ef53b958ef03d279232007ad3319b1f38584bdb
    - docker.io/cloudbees/cloudbees-core-oc:2.289.2.2@sha256:9cd85ee84e401dc27e3a8268aae67b594a651b2f4c7fc056ca14c7b0a0a6b82d
    - docker.io/anchore/test_images:grype-quality-node-d89207b@sha256:f56164678054e5eb59ab838367373a49df723b324617b1ba6de775749d7f91d4

  partial-label-set-images: &partial-label-set-images
    - docker.io/vulhub/cve-2017-1000353:latest@sha256:da2a59314b9ccfb428a313a7f163adcef77a74a393b8ebadeca8223b8cea9797

result-sets:

  sboms:

    description: "SBOMs for images that should be fully labeled"
    matrix:
      images:
        - *full-label-set-images
        - *partial-label-set-images

      tools:

        - name: syft
          # note: we want to use a fixed version of syft for capturing all results (NOT "latest")
          version: v0.68.1
          # once we have results captured, don't re-capture them
          refresh: false
"""
    file = tmp_path / "config.yaml"
    file.write_text(subject)

    profile_text = """
test_profile:
  something:
    name: jim
    config_path: .abc/xyx.conf
    refresh: false
"""
    profile_file.write_text(profile_text)

    cfg = config.load(str(file))

    assert cfg.result_sets["sboms"].matrix.images == [
        "docker.io/cloudbees/cloudbees-core-agent:2.289.2.2@sha256:d48f0546b4cf5ef4626136242ce302f94a42751156b7be42f4b1b75a66608880",
        "docker.io/cloudbees/cloudbees-core-mm:2.277.3.1@sha256:4c564f473d38f23da1caa48c4ef53b958ef03d279232007ad3319b1f38584bdb",
        "docker.io/cloudbees/cloudbees-core-oc:2.289.2.2@sha256:9cd85ee84e401dc27e3a8268aae67b594a651b2f4c7fc056ca14c7b0a0a6b82d",
        "docker.io/anchore/test_images:grype-quality-node-d89207b@sha256:f56164678054e5eb59ab838367373a49df723b324617b1ba6de775749d7f91d4",
        "docker.io/vulhub/cve-2017-1000353:latest@sha256:da2a59314b9ccfb428a313a7f163adcef77a74a393b8ebadeca8223b8cea9797",
    ]

    assert cfg.profiles == config.Profiles(
        {
            "test_profile": {
                "something": {
                    "name": "jim",
                    "config_path": ".abc/xyx.conf",
                    "refresh": False,
                },
            },
        },
    )


@pytest.mark.parametrize(
    "name, image, expected_valid",
    [
        # valid: everything present
        (
            "valid",
            "registry.example.com/repo/image:latest@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            True,
        ),
        (
            "valid: vulhub",
            "docker.io/vulhub/cve-2017-1000353:latest@sha256:da2a59314b9ccfb428a313a7f163adcef77a74a393b8ebadeca8223b8cea9797",
            True,
        ),
        (
            "valid: alpine",
            "docker.io/alpine:3.2@sha256:ddac200f3ebc9902fb8cfcd599f41feb2151f1118929da21bcef57dc276975f9",
            True,
        ),
        # valid: localhost with port as repo host
        (
            "valid: localhost with port as repo host",
            "localhost:5555/repo/image:latest@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            True,
        ),
        (
            "valid: missing tag is allowed but discouraged",
            "registry.access.redhat.com/ubi8@sha256:68fecea0d255ee253acbf0c860eaebb7017ef5ef007c25bee9eeffd29ce85b29",
            True,
        ),
        (
            "invalid: missing host",
            "repo/image:latest@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            False,
        ),
        ("invalid: missing digest", "registry.example.com/repo/image:latest", False),
        ("invalid: missing everything", "repo/image", False),
        ("invalid: empty string", "", False),
        (
            "invalid: missing repo",
            "registry.example.com/:latest@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            False,
        ),
        ("invalid: missing repo and tag", "registry.example.com/", False),
        ("invalid: missing digest", "registry.example.com/repo/image:stable", False),
        (
            "invalid: digest does not look like sha256",
            "registry.example.com/repo/image:latest@sha256:invaliddigest",
            False,
        ),
        (
            "invalid: bad sha256 (too short)",
            "docker.io/alpine:3.2@sha256:ddac200f3ebc9902fb8cfcd599f41feb2151f1118929da21bcef57dc27697",
            False,
        ),
    ],
)
def test_is_valid_oci_reference(name, image, expected_valid):
    result = config.ScanMatrix.is_valid_oci_reference(image)
    assert result == expected_valid, f"Test case {name}: Expected {expected_valid} but got {result} for image '{image}'"


@pytest.mark.parametrize(
    "image, expected_output",
    [
        (
            "docker.io/anchore/test_images:some-tag@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            (
                "docker.io",
                "anchore",
                "test_images",
                "some-tag",
                "sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            ),
        ),
        # Localhost reference with path, repository, tag, and digest
        (
            "localhost/anchore/test_images:some-tag@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            (
                "localhost",
                "anchore",
                "test_images",
                "some-tag",
                "sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            ),
        ),
        # Localhost with port, path, repository, tag, and digest
        (
            "localhost:5000/anchore/test_images:some-tag@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            (
                "localhost:5000",
                "anchore",
                "test_images",
                "some-tag",
                "sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            ),
        ),
        # Missing digest
        (
            "docker.io/anchore/test_images:some-tag",
            ("docker.io", "anchore", "test_images", "some-tag", ""),
        ),
        # Missing tag
        (
            "docker.io/anchore/test_images@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            (
                "docker.io",
                "anchore",
                "test_images",
                "",
                "sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef",
            ),
        ),
        # Only repository
        ("test_images", ("", "", "test_images", "", "")),
    ],
)
def test_parse_oci_reference(image, expected_output):
    result = config.ScanMatrix.parse_oci_reference(image)
    assert result == expected_output, f"Expected {expected_output} but got {result} for image '{image}'"
